﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace NMag.Core
{
    public enum ModuleType
    {
        Reservoir = 0,
        PowerPlant = 1,
        Transfer = 2,
        ControlPoint = 3,
        Combined = 4
    }

    public abstract class Module
    {
        private EnMagOperationalStrategy _enMagOperationalStrategy = new EnMagOperationalStrategy();
        private NMagOperationalStrategy _nMagOperationalStrategy = new NMagOperationalStrategy();

        public Module(ModuleType type, int number)
        {
            ModuleType = type;
            Number = number;
        }
        
        public Module(Module module)
        {
            ModuleType = module.ModuleType;
            Number = module.Number;

            ReleaseModuleNumber = module.ReleaseModuleNumber;
            BypassModuleNumber = module.BypassModuleNumber;
            SpillModuleNumber = module.SpillModuleNumber;

            EnableRoutingCurve = module.EnableRoutingCurve;
            RoutingCurve = new RoutingCurve(module.RoutingCurve);

            UseLocalInflow = module.UseLocalInflow;
            AnnualRunoff = module.AnnualRunoff;

            RunoffSeriesWeights = new Dictionary<string, double>();
            foreach (KeyValuePair<string, double> weight in module.RunoffSeriesWeights)
            {
                RunoffSeriesWeights.Add(weight.Key, weight.Value);
            }

            UseMinimumFlow = module.UseMinimumFlow;
            MinimumFlowCurve = new MinimumFlowCurve(module.MinimumFlowCurve);

            UseBypassRelease = module.UseBypassRelease;
            BypassReleaseCurve = new BypassReleaseCurve(module.BypassReleaseCurve);

            UseMinimumLevel = module.UseMinimumLevel;
            MinimumLevelCurve = new MinimumLevelCurve(module.MinimumLevelCurve);

            UseMaximumLevel = module.UseMaximumLevel;
            MaximumLevelCurve = new MaximumLevelCurve(module.MaximumLevelCurve);

            UseOperationalStrategy = module.UseOperationalStrategy;
            _enMagOperationalStrategy = module._enMagOperationalStrategy?.Clone();
            _nMagOperationalStrategy = module._nMagOperationalStrategy?.Clone();
        }

        public ModuleType ModuleType { get; protected set; }
        public string TypeString
        {
            get
            {
                return string.Join("", ModuleType
                    .ToString()
                    .Select(c => (char.IsUpper(c)) ? " " + c.ToString() : c.ToString()))
                    .Trim();
            }        
        }

        public string Name
        {
            get
            {
                List<string> unitNames = new List<string>();
                unitNames.AddRange(UnitHelper.GetReservoirs(Units).Select(u => u.Name));
                unitNames.AddRange(UnitHelper.GetPowerPlants(Units).Select(u => u.Name));
                unitNames.AddRange(UnitHelper.GetTransfers(Units).Select(u => u.Name));
                unitNames.AddRange(UnitHelper.GetControlPoints(Units).Select(u => u.Name));

                return string.Join("/", unitNames) ?? "";
            }
        }
        public int Number { get; set; }

        public List<Unit> Units { get; private set; } = new List<Unit>();

        public int ReleaseModuleNumber { get; set; } = 0;
        public int BypassModuleNumber { get; set; } = 0;
        public int SpillModuleNumber { get; set; } = 0;

        public int[] RoutingNumbers
        {
            get
            {
                return new int[]
                {
                    ReleaseModuleNumber,
                    BypassModuleNumber,
                    SpillModuleNumber
                };
            }
        }

        public bool EnableRoutingCurve { get; set; }
        public RoutingCurve RoutingCurve { get; set; } = new RoutingCurve(new List<Point>());

        public bool UseLocalInflow { get; set; } = false;
        public double AnnualRunoff { get; set; }
        public Dictionary<string, double> RunoffSeriesWeights { get; set; } = new Dictionary<string, double>();

        public bool UsesRestrictions
        {
            get { return UseMinimumFlow || UseBypassRelease || UseMinimumLevel || UseMaximumLevel; }
        }

        public bool UseMinimumFlow { get; set; }
        public MinimumFlowCurve MinimumFlowCurve { get; set; } = new MinimumFlowCurve(new List<Point>());           // Curve code 100

        public bool UseBypassRelease { get; set; }
        public BypassReleaseCurve BypassReleaseCurve { get; set; } = new BypassReleaseCurve(new List<Point>());     // Curve code 200

        public bool UseMinimumLevel { get; set; }
        public MinimumLevelCurve MinimumLevelCurve { get; set; } = new MinimumLevelCurve(new List<Point>());        // Curve code 300

        public bool UseMaximumLevel { get; set; }
        public MaximumLevelCurve MaximumLevelCurve { get; set; } = new MaximumLevelCurve(new List<Point>());        // Curve code 400

        public bool UseOperationalStrategy { get; set; } = false;
        public EnMagOperationalStrategy EnMagOperationalStrategy
        {
            get { return _enMagOperationalStrategy; }
            set
            {
                if (value != null)
                {
                    _enMagOperationalStrategy = value;
                    UseOperationalStrategy = true;
                }
            }
        }
        public NMagOperationalStrategy NMagOperationalStrategy
        {
            get { return _nMagOperationalStrategy; }
            set
            {
                if (value != null)
                {
                    _nMagOperationalStrategy = value;
                    UseOperationalStrategy = true;
                }
            }
        }

        public abstract Module Clone();

        public void AddRunoffSeries(string name)
        {
            RunoffSeriesWeights.Add(name, 0);
        }
        public void RenameRunoffSeries(string oldName, string newName)
        {
            RunoffSeriesWeights.Add(newName, RunoffSeriesWeights[oldName]);
            RunoffSeriesWeights.Remove(oldName);
        }
        public void RemoveRunoffSeries(string name)
        {
            RunoffSeriesWeights.Remove(name);
        }

        public Result Validate()
        {
            if (!Enum.IsDefined(typeof(ModuleType), ModuleType)) { return new Result(false, "Invalid module type."); }

            return new Result(true);
        }
    }
}
