﻿using NMag.Core.Utilization;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;

namespace NMag.Core
{
    public class DataSetReader
    {
        private Project _project;
        private bool _enMagMode;

        private List<Module> _modules = new List<Module>();

        private List<string> _lines = new List<string>();
        private int _lineIndex = 0;

        private int _failureIndex = -1;
        private bool _readingFailed = false;
        private bool ReadingFailed
        {
            get { return _readingFailed; }
            set
            {
                if(value)
                {
                    _readingFailed = true;
                    _failureIndex = _lineIndex + 1;
                }
            }
        }

        /// <summary>
        /// Creates a project using the data from a given data set.
        /// </summary>
        /// <param name="filePath">Path of the data set.</param>
        /// <param name="project">Project created from the data set.</param>
        /// <returns></returns>
        public Result Read(string filePath, out Project project)
        {
            project = null;
            ClearReaderData();

            Result result = LoadLines(filePath);
            if(!result.Succes) { return result; }

            _project = new Project(Path.GetFileNameWithoutExtension(filePath));

            result = ReadLines();
            if(!result.Succes) { return result; }

            result = _project.CreateModules(_modules);

            project = _project;
            return new Result(true, result.Message == "" ? "Data set successfully read." : result.Message);
        }

        /// <summary>
        /// Clears reader data
        /// </summary>
        private void ClearReaderData()
        {
            _modules.Clear();
            _lines.Clear();
            _lineIndex = 0;

            _failureIndex = -1;
            _readingFailed = false;
        }

        /// <summary>
        /// Loads the data from a data set file.
        /// </summary>
        /// <param name="filePath">Path of the data set.</param>
        /// <returns></returns>
        private Result LoadLines(string filePath)
        {
            if (Path.GetExtension(filePath).ToLower() != PathFinder.GetExtension(PathFinder.FileType.DataSet))
            {
                return new Result(false, "Selected file is not a data set.");
            }

            try { _lines = File.ReadAllLines(filePath, Encoding.UTF7).ToList(); }
            catch { return new Result(false, "Could not open data set."); }

            return new Result(true);
        }
       
        /// <summary>
        /// Reads the data set lines and stores the data in the project.
        /// </summary>
        /// <returns></returns>
        private Result ReadLines()
        {
            UnpackCounters();

            while(!EndOfLines() && !ReadingFailed)
            {
                if(GetLine()[0] == '*')
                {
                    switch(GetLine().Substring(0, 3))
                    {
                        // EnMag mode
                        case "*MO":
                            if(!UnpackEnMagMode())
                            {
                                return new Result(false, $"Reading data set failed at EnMag Mode (*MO) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Comments
                        case "*CO":
                            if (!UnpackComments())
                            {
                                return new Result(false, $"Reading data set failed at Comments (*CO) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Reservoir
                        case "*MA":
                            if (!UnpackReservoirs())
                            {
                                return new Result(false, $"Reading data set failed at Reservoirs (*MA) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Power plant
                        case "*KR":
                            if (!UnpackPowerPlants())
                            {
                                return new Result(false, $"Reading data set failed at Power Plants (*KR) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Transfer
                        case "*OF":
                            if (!UnpackTransfers())
                            {
                                return new Result(false, $"Reading data set failed at Transfers (*OF) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Control point
                        case "*KO":
                            if (!UnpackControlPoints())
                            {
                                return new Result(false, $"Reading data set failed at Control Points (*KO) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Energy market
                        case "*KM":
                            if (!UnpackEnergyMarket())
                            {
                                return new Result(false, $"Reading data set failed at Energy Market (*KM) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Restrictions
                        case "*RS":
                            if (!UnpackRestrictions())
                            {
                                return new Result(false, $"Reading data set failed at Restrictions (*RS) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Operational strategy
                        case "*ST":
                            if(_enMagMode)
                            {
                                if (!UnpackEnMagOperationalStrategy())
                                {
                                    return new Result(false, $"Reading data set failed at Operational Strategy (*ST) at line {_failureIndex}.");
                                }
                            }
                            else
                            {
                                if (!UnpackNMagOperationalStrategy())
                                {
                                    return new Result(false, $"Reading data set failed at Operational Strategy (*ST) at line {_failureIndex}.");
                                }
                            }
                            _lineIndex++;
                            break;

                        // Hydrology
                        case "*HY":
                            if (!UnpackHydrology())
                            {
                                return new Result(false, $"Reading data set failed at Hydrology (*HY) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        // Job control
                        case "*JK":
                            if (!UnpackJobControl())
                            {
                                return new Result(false, $"Reading data set failed at Job Control (*JK) at line {_failureIndex}.");
                            }
                            _lineIndex++;
                            break;

                        case "***":
                        case "*EX":
                            return new Result(true, "All lines read successfully.");

                        default:
                            return new Result(false, $"Error reading data set on line {_failureIndex + 1}.");
                    }
                }
                else
                {
                    _lineIndex++;
                }
            }

            return new Result(true, "All lines read successfully.");
        }

        /// <summary>
        /// Returns the current line.
        /// Returns empty string and sets Reading failed to true when end of data set is reached.
        /// </summary>
        /// <returns></returns>
        private string GetLine()
        {
            if(EndOfLines())
            {
                ReadingFailed = true;
                return "";
            }
            
            return _lines[_lineIndex];
        }

        /// <summary>
        /// Returns true when the end of the data set is reached.
        /// </summary>
        /// <returns></returns>
        private bool EndOfLines()
        {
            return _lineIndex >= _lines.Count;
        }

        /// <summary>
        /// Returns first value of current line parsed to an int.
        /// Sets next line as current line.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <returns></returns>
        private int ReadSingleInt()
        {
            string text = GetLine().Split(',').First();

            ReadingFailed = !int.TryParse(text, out int number);

            _lineIndex++;

            return number;
        }

        /// <summary>
        /// Returns first value of given text parsed to an int.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <param name="text">Text to be parsed.</param>
        /// <returns></returns>
        private int ReadSingleInt(string text)
        {
            text = text.Split(',').FirstOrDefault();

            ReadingFailed = !int.TryParse(text, out int number);

            return number;
        }

        /// <summary>
        /// Returns values of current line parsed to int.
        /// Sets next line as current line.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <param name="correctCount">Expected count of values. Mismatch will result in ReadingFailed.</param>
        /// <returns></returns>
        private List<int> ReadMultipleInt(int correctCount = 0)
        {
            List<int> values = GetLine()
                .Split(',')
                .Where(v => int.TryParse(v, out int parsedValue))
                .Select(v => int.Parse(v)).ToList();

            if((correctCount > 0) && (values.Count != correctCount))
            {
                ReadingFailed = true;
            }

            _lineIndex++;

            return values;
        }

        /// <summary>
        /// Returns first value of current line parsed to a double.
        /// Sets next line as current line.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <returns></returns>
        private double ReadSingleDouble()
        {
            string text = GetLine().Split(',').First();

            ReadingFailed = !double.TryParse(text, out double number);

            _lineIndex++;

            return number;
        }

        /// <summary>
        /// Returns first value of given text parsed to a double.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <param name="text">Text to be parsed.</param>
        /// <returns></returns>
        private double ReadSingleDouble(string text)
        {
            ReadingFailed = !double.TryParse(text, out double number);

            return number;
        }

        /// <summary>
        /// Returns values of current line parsed to double.
        /// Sets next line as current line.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <param name="correctCount">Expected count of values. Mismatch will result in ReadingFailed.</param>
        /// <returns></returns>
        private List<double> ReadMultipleDouble(int correctCount = 0)
        {
            List<double> values = GetLine()
                .Split(',')
                .Where(v => double.TryParse(v, out double parsedValue))
                .Select(v => double.Parse(v)).ToList();

            if ((correctCount > 0) && (values.Count != correctCount))
            {
                ReadingFailed = true;
            }

            _lineIndex++;

            return values;
        }

        /// <summary>
        /// Returns first value of current line as string.
        /// Sets next line as current line.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <returns></returns>
        private string ReadSingleString()
        {
            string text = GetLine().Split(',').First();

            ReadingFailed = EndOfLines();

            _lineIndex++;

            return text;
        }

        /// <summary>
        /// Returns values of current line as string.
        /// Sets next line as current line.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <param name="correctCount">Expected count of values. Mismatch will result in ReadingFailed.</param>
        /// <returns></returns>
        private List<string> ReadMultipleString(int correctCount = 0)
        {
            List<string> text = GetLine().Split(',').ToList();
            text.RemoveAt(text.Count - 1);

            if ((correctCount > 0) && (text.Count != correctCount))
            {
                ReadingFailed = true;
            }

            _lineIndex++;

            return text;
        }

        /// <summary>
        /// Returns a collection of lines from current line until a * or the end of the data set.
        /// </summary>
        /// <returns></returns>
        private IEnumerable<string> ReadMultipleLines()
        {
            while ((GetLine()[0] != '*') && !EndOfLines())
            {
                yield return GetLine();
                _lineIndex++;
            }
        }

        /// <summary>
        /// Returns a parsed list of points, starting from current line.
        /// Sets ReadingFailed to true on error.
        /// </summary>
        /// <param name="pointCount">The number of lines that must be read.</param>
        /// <returns></returns>
        private List<Point> ReadCurve(int pointCount)
        {
            List<Point> points = new List<Point>();

            for (int i = 0; i < pointCount; i++)
            {
                List<double> coordinates = ReadMultipleDouble(2);

                if(ReadingFailed)
                {
                    return new List<Point>();
                }
                
                points.Add(new Point(coordinates[0], coordinates[1]));
            }

            return points;
        }

        /// <summary>
        /// Returns the next 365 values as a curve starting from current line.
        /// Sets ReadingFailed to true on error.
        /// Will set error to true when number of values is not 365.
        /// </summary>
        /// <returns></returns>
        private List<Point> ReadAnnualCurve()
        {
            int dayCount = 0;
            List<Point> points = new List<Point>();

            do
            {
                List<double> dayValues = ReadMultipleDouble();

                if(ReadingFailed) { return new List<Point>(); }

                foreach(double dayValue in dayValues)
                {
                    dayCount++;
                    points.Add(new Point(dayCount, dayValue));
                }
            } while ((dayCount < 365));

            if(dayCount > 365)
            {
                ReadingFailed = true;
                return new List<Point>();
            }

            return points;
        }

        /// <summary>
        /// Returns true when the current line contains one of the given curve codes.
        /// </summary>
        /// <param name="curveCodes">List of allowed curve codes.</param>
        /// <returns></returns>
        private bool IsCurveCode(int[] curveCodes)
        {
            if(GetLine().Count(c => c == ',') != 1)
            {
                return false;
            }

            if(!int.TryParse(GetLine().DigitsOnly(), out int code))
            {
                return false;
            }

            if(!curveCodes.Contains(code))
            {
                return false;   
            }

            return true;
        }

        /// <summary>
        /// Skips first 7 lines.
        /// This information is obsolete.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackCounters()
        {
            _lineIndex += 6;

            _project.ProjectSettings.MaxCurv = ReadSingleInt();

            if(EndOfLines())
            {
                ReadingFailed = true; 
            }

            return !ReadingFailed;
        }

        /// <summary>
        /// Sets the EnMag mode of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackEnMagMode()
        {
            _lineIndex++;

            _enMagMode = ReadSingleInt() == 1 ? true : false;

            return !ReadingFailed;
        }

        /// <summary>
        /// Sets the comments of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackComments()
        {
            _lineIndex++;

            _project.ProjectSettings.Comments = ReadMultipleLines().ToList();

            return !ReadingFailed;
        }

        /// <summary>
        /// Unpacks reservoir modules.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackReservoirs()
        {
            _lineIndex++;

            if(GetLine()[0] == '*' || EndOfLines())
            {
                return true;
            }

            while (GetLine()[0] != '*' && !EndOfLines() && !ReadingFailed)
            {
                int moduleNumber = ReadSingleInt();
                string unitName = ReadSingleString();

                double lrwl = ReadSingleDouble();
                double hrwl = ReadSingleDouble();
                double volume = ReadSingleDouble();

                List<int> routing = ReadMultipleInt(3);
                if(ReadingFailed)
                {
                    return false;
                }

                ReservoirModule reservoirModule = new ReservoirModule(moduleNumber, unitName, lrwl, hrwl, volume)
                {
                    ReleaseModuleNumber = routing[0],
                    BypassModuleNumber = routing[1],
                    SpillModuleNumber = routing[2]
                };

                int[] curveCodes = { 200, 300, 400, 500, 600 };

                while (GetLine()[0] != '*' && !EndOfLines() && !ReadingFailed)
                {
                    // Check if current line has a curve code
                    if (!IsCurveCode(curveCodes))
                    {
                        break;
                    }

                    int curveCode = ReadSingleInt();

                    switch(curveCode)
                    {
                        case 200:
                            reservoirModule.Reservoir.UseVolumeCurve = true;
                            reservoirModule.Reservoir.VolumeCurve = new VolumeCurve(ReadCurve(ReadSingleInt()));
                            break;

                        case 300:
                            reservoirModule.Reservoir.UseSpillFlowCurve = true;
                            reservoirModule.Reservoir.SpillFlowCurve = new SpillFlowCurve(ReadCurve(ReadSingleInt()));
                            break;

                        case 400:
                            reservoirModule.Reservoir.UseGateDischargeCurve = true;
                            reservoirModule.Reservoir.OutletLevel = ReadSingleDouble();
                            reservoirModule.Reservoir.GateDischargeCurve = new GateDischargeCurve(ReadCurve(ReadSingleInt()));
                            break;

                        case 500:
                            reservoirModule.Reservoir.UseEvaporationCurve = true;
                            reservoirModule.Reservoir.EvaporationCurve = new EvaporationCurve(ReadCurve(ReadSingleInt()));
                            break;
                            
                        case 600:
                            reservoirModule.EnableRoutingCurve = true;
                            reservoirModule.RoutingCurve = new RoutingCurve(ReadCurve(ReadSingleInt()));
                            break;

                        default:
                            ReadingFailed = true;
                            return false;
                    }
                }

                if(ReadingFailed) { return false; }

                _modules.Add(reservoirModule);
            }

            return true;
        }

        /// <summary>
        /// Sets the power plants of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackPowerPlants()
        {
            _lineIndex++;

            if (GetLine()[0] == '*' || EndOfLines())
            {
                return true;
            }

            while((GetLine()[0] != '*') && !EndOfLines() && !ReadingFailed)
            {
                int moduleNumber = ReadSingleInt();
                string unitName = ReadSingleString();

                List<int> routing = ReadMultipleInt(3);
                if(ReadingFailed)
                {
                    return false;
                }
                
                double maxDischarge = ReadSingleDouble();
                double energyEquivalent = ReadSingleDouble();

                PowerPlantModule powerPlantModule = new PowerPlantModule(moduleNumber, unitName, maxDischarge, energyEquivalent)
                {
                    ReleaseModuleNumber = routing[0],
                    BypassModuleNumber = routing[1],
                    SpillModuleNumber = routing[2]
                };

                int[] curveCodes = { 200, 400, 600 };

                while (GetLine()[0] != '*' && !EndOfLines() && !ReadingFailed)
                {
                    // Check if current line has a curve code
                    if (!IsCurveCode(curveCodes))
                    {
                        break;
                    }

                    int curveCode = ReadSingleInt();

                    switch (curveCode)
                    {
                        case 200:
                            powerPlantModule.PowerPlant.UseOptionalData = true;

                            powerPlantModule.PowerPlant.NominalHead = ReadSingleDouble();
                            powerPlantModule.PowerPlant.IntakeLevel = ReadSingleDouble();
                            powerPlantModule.PowerPlant.TailWaterLevel = ReadSingleDouble();
                            powerPlantModule.PowerPlant.HeadLossCoefficient = ReadSingleDouble();

                            powerPlantModule.PowerPlant.LoadEfficiencyCurve = new LoadEfficiencyCurve(ReadCurve(ReadSingleInt()));
                            break;
     
                        case 400:
                            powerPlantModule.PowerPlant.UsePeakingCurve = true;
                            powerPlantModule.PowerPlant.PeakingCurve = new PeakingCurve(ReadCurve(ReadSingleInt()));
                            break;

                        case 600:
                            powerPlantModule.EnableRoutingCurve = true;
                            powerPlantModule.RoutingCurve = new RoutingCurve(ReadCurve(ReadSingleInt()));
                            break;

                        default:
                            ReadingFailed = true;
                            return false;
                    }
                }

                if (ReadingFailed) { return false; }

                _modules.Add(powerPlantModule);
            }

            return true;
        }

        /// <summary>
        /// Sets transfers of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackTransfers()
        {
            _lineIndex++;

            if (GetLine()[0] == '*' || EndOfLines())
            {
                return true;
            }

            while (GetLine()[0] != '*' && !EndOfLines() && !ReadingFailed)
            {
                int moduleNumber = ReadSingleInt();
                string unitName = ReadSingleString();

                List<int> routing = ReadMultipleInt(3);
                if(ReadingFailed)
                {
                    return  false;
                }

                TransferModule transferModule = new TransferModule(moduleNumber, unitName)
                {
                    ReleaseModuleNumber = routing[0],
                    BypassModuleNumber = routing[1],
                    SpillModuleNumber = routing[2]
                };

                int[] curveCodes = { 100, 200, 300, 600 };

                while (GetLine()[0] != '*' && !EndOfLines() && !ReadingFailed)
                {
                    // Check if current line has a curve code
                    if (!IsCurveCode(curveCodes))
                    {
                        break;
                    }

                    int curveCode = ReadSingleInt();

                    switch (curveCode)
                    {
                        case 100:
                            transferModule.Transfer.Mode = TransferUnit.TransferMode.FixedCapacity;
                            transferModule.Transfer.Capacity = ReadSingleDouble();
                            break;

                        case 200:
                            transferModule.Transfer.Mode = TransferUnit.TransferMode.VariableCapacity;

                            List<double> values = ReadMultipleDouble(3);
                            if (ReadingFailed) { return false; }
                            transferModule.Transfer.InletElevation = values[0];
                            transferModule.Transfer.OutletElevation = values[1];
                            transferModule.Transfer.HeadLossCoefficient = values[2];
                            break;

                        case 300:
                            transferModule.Transfer.Mode = TransferUnit.TransferMode.HQCapacityCurve;

                            transferModule.Transfer.IntakeElevation = ReadSingleDouble();
                            transferModule.Transfer.TailwaterElevation = ReadSingleDouble();

                            transferModule.Transfer.HQCapacityCurve = new HQCapacityCurve(ReadCurve(ReadSingleInt()));
                            break;

                        case 600:
                            transferModule.EnableRoutingCurve = true;
                            transferModule.RoutingCurve = new RoutingCurve(ReadCurve(ReadSingleInt()));
                            break;

                        default:
                            ReadingFailed = true;
                            return false;
                    }
                }

                if (ReadingFailed) { return false; }

                _modules.Add(transferModule);
            }
            return true;
        }

        /// <summary>
        /// Sets control points of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackControlPoints()
        {
            _lineIndex++;

            if (GetLine()[0] == '*' || EndOfLines())
            {
                return true;
            }

            while (GetLine()[0] != '*' && !EndOfLines() && !ReadingFailed)
            {
                int moduleNumber = ReadSingleInt();
                string unitName = ReadSingleString();

                List<int> routing = ReadMultipleInt(3);
                if(ReadingFailed)
                {
                    return false;   
                }

                ControlPointModule controlPointModule = new ControlPointModule(moduleNumber, unitName)
                {
                    ReleaseModuleNumber = routing[0],
                    BypassModuleNumber = routing[1],
                    SpillModuleNumber = routing[2]
                };

                if(GetLine() == "600,")
                {
                    _lineIndex++;
                    controlPointModule.EnableRoutingCurve = true;
                    controlPointModule.RoutingCurve = new RoutingCurve(ReadCurve(ReadSingleInt()));
                }

                if (ReadingFailed)
                {
                    return false;
                }

                _modules.Add(controlPointModule);
            }

            return true;
        }

        /// <summary>
        /// Sets the energy market of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackEnergyMarket()
        {
            _lineIndex++;

            EnergyMarket energyMarket = new EnergyMarket()
            {
                FirmPowerLevel = ReadSingleDouble(),
                FirmPowerPrice = ReadSingleDouble()
            };

            List<int> summerPeriod = ReadMultipleInt(2);
            if(ReadingFailed)
            {
                return false;
            }

            if((summerPeriod[0] < 0) 
            || (summerPeriod[0] > _project.ProjectSettings.TimeResolution)
            || (summerPeriod[0] >= summerPeriod[1])
            || (summerPeriod[1] < 0) 
            || (summerPeriod[1] > _project.ProjectSettings.TimeResolution))
            {
                ReadingFailed = true;
                return false;
            }

            energyMarket.SummerStart = summerPeriod[0];
            energyMarket.SummerEnd = summerPeriod[1];

            int distributionMode = ReadSingleInt();

            switch(distributionMode)
            {
                case 1:
                    energyMarket.Mode = EnergyMarket.DistributionMode.DailyLoad;
                    energyMarket.DailyLoadCurve = new DailyLoadCurve(ReadAnnualCurve());
                    break;

                case 3:
                    energyMarket.Mode = EnergyMarket.DistributionMode.Vairable;
                    energyMarket.VariableLoadCurve = new VariableLoadCurve(ReadCurve(ReadSingleInt()));
                    break;

                // No data for this mode
                case 4:
                    break;

                default:
                    ReadingFailed = true;
                    return false;
            }

            if (GetLine()[0] == '*' || EndOfLines())
            {
                return true;
            }

            int preferenceFunctionCount = ReadSingleInt();

            for(int i = 0; i < preferenceFunctionCount; i++)
            {
                // Skip preference number
                ReadSingleInt();

                List<int> period = ReadMultipleInt(2);
                if(ReadingFailed)
                {
                    return false;
                }

                if((period[0] < 0)
                || (period[0] > _project.ProjectSettings.TimeResolution)
                || (period[0] >= period[1])
                || (period[1] < 0)
                || (period[1] > _project.ProjectSettings.TimeResolution))
                {
                    ReadingFailed = true;
                    return false;
                }

                PreferenceCurve preferenceCurve = new PreferenceCurve(ReadCurve(ReadSingleInt()));
                preferenceCurve.ConvertToContinuous();

                PreferenceFunction function = new PreferenceFunction()
                {
                    Start = period[0],
                    End = period[1],
                    PreferenceCurve = preferenceCurve
                };

                List<int> steps = ReadMultipleInt(2);
                if(ReadingFailed)
                {
                    return false;
                }

                function.RationingStart = steps[0];
                function.SellingStart = steps[1];

                energyMarket.PreferenceFunctions.Add(function);
            }

            if(ReadingFailed)
            {
                return false;
            }

            _project.EnergyMarket = energyMarket;
            return true;
        }

        /// <summary>
        /// Sets the restrictions of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackRestrictions()
        {
            _lineIndex++;

            if (GetLine()[0] == '*' || EndOfLines())
            {
                return true;
            }

            int[] curveNumbers = { 100, 200, 300, 400 };

            do
            {
                int curveCode = ReadSingleInt();

                if(!curveNumbers.Contains(curveCode))
                {
                    ReadingFailed = true;
                    return false;
                }

                int moduleNumber = ReadSingleInt();

                switch(curveCode)
                {
                    case 100:
                        MinimumFlowCurve minimumFlowCurve = new MinimumFlowCurve(ReadCurve(ReadSingleInt()));
                        foreach(Module module in _modules.Where(m => m.Number == moduleNumber))
                        {
                            module.UseMinimumFlow = true;
                            module.MinimumFlowCurve = minimumFlowCurve;
                        }
                        break;

                    case 200:
                        BypassReleaseCurve bypassCurve = new BypassReleaseCurve(ReadCurve(ReadSingleInt()));
                        foreach (Module module in _modules.Where(m => m.Number == moduleNumber))
                        {
                            module.UseBypassRelease = true;
                            module.BypassReleaseCurve = bypassCurve;
                        }
                        break;

                    case 300:
                        MinimumLevelCurve minimumLevelCure = new MinimumLevelCurve(ReadCurve(ReadSingleInt()));
                        foreach (Module module in _modules.Where(m => m.Number == moduleNumber))
                        {
                            module.UseMinimumLevel = true;
                            module.MinimumLevelCurve = minimumLevelCure;
                        }
                        break;

                    case 400:
                        MaximumLevelCurve MaximumLevelCure = new MaximumLevelCurve(ReadCurve(ReadSingleInt()));
                        foreach (Module module in _modules.Where(m => m.Number == moduleNumber))
                        {
                            module.UseMaximumLevel = true;
                            module.MaximumLevelCurve = MaximumLevelCure;
                        }
                        break;

                    default:
                        ReadingFailed = true;
                        return false;
                }

            } while (GetLine()[0] != '*');

            if (ReadingFailed) { return false; }
            
            return true;
        }

        /// <summary>
        /// Sets EnMag operational strategy of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackEnMagOperationalStrategy()
        {
            _lineIndex++;

            while (GetLine()[0] != '*' && !EndOfLines())
            {
                int strategyCode = ReadSingleInt();

                int moduleNumber = ReadSingleInt();

                EnMagOperationalStrategy strategy = new EnMagOperationalStrategy();

                switch(strategyCode)
                {
                    case 300:
                        strategy.Strategy = EnMagOperationalStrategy.OperationalStrategy.Unconditional;
                        break;

                    case 500:
                        strategy.Strategy = EnMagOperationalStrategy.OperationalStrategy.GuideCurve;
                        strategy.GuideCurve = new GuideCurve(ReadCurve(ReadSingleInt()));
                        break;

                    case 600:
                        strategy.Strategy = EnMagOperationalStrategy.OperationalStrategy.RuleCurve;
                        strategy.RationingRuleCurve = new RationingRuleCurve(ReadCurve(ReadSingleInt()));
                        strategy.SurplusRuleCurve = new SurplusRuleCurve(ReadCurve(ReadSingleInt()));
                        break;

                    default:
                        ReadingFailed = true;
                        return false;
                }

                if(ReadingFailed)
                {
                    return false;
                }

                _modules.Where(m => m.Number == moduleNumber)
                    .ToList()
                    .ForEach(m => m.EnMagOperationalStrategy = strategy);
            }

            return true;
        }

        /// <summary>
        /// Sets nMag operational strategy of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackNMagOperationalStrategy()
        {
            _lineIndex++;

            while (GetLine()[0] != '*' && !EndOfLines())
            {
                int strategyCode = ReadSingleInt();

                List<double> values = ReadMultipleDouble();
                if(ReadingFailed)
                {
                    return false;
                }

                int moduleNumber = Convert.ToInt32(values[0]);

                NMagOperationalStrategy strategy = new NMagOperationalStrategy();

                switch (strategyCode)
                {
                    case 300:
                        strategy.Strategy = NMagOperationalStrategy.OperationalStrategy.Automatic;
                        break;

                    case 400:
                        strategy.Strategy = NMagOperationalStrategy.OperationalStrategy.ReservoirRelease;

                        if(values.Count == 1)
                        {
                            ReadingFailed = true;
                            return false;
                        }

                        strategy.AnnualRelease = values[1];

                        strategy.ReservoirReleaseCurve = new ReservoirReleaseCurve(ReadCurve(ReadSingleInt()));
                        break;

                    case 500:
                        strategy.Strategy = NMagOperationalStrategy.OperationalStrategy.ReservoirGuide;
                        strategy.ReservoirGuideCurve = new ReservoirGuideCurve(ReadCurve(ReadSingleInt()));
                        break;

                    default:
                        ReadingFailed = true;
                        return false;
                }

                if (ReadingFailed) { return false; }

                _modules.Where(m => m.Number == moduleNumber)
                    .ToList()
                    .ForEach(m => m.NMagOperationalStrategy = strategy);
            }

            return true;
        }

        /// <summary>
        /// Sets hydrology of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackHydrology()
        {
            _lineIndex++;

            Hydrology hydrology = new Hydrology()
            {
                TimeResolution = _project.ProjectSettings.TimeResolution
            };

            List<int> period = ReadMultipleInt(2);
            if(ReadingFailed)
            {
                return false;
            }

            int moduleHydrologyCount = ReadSingleInt();
            int runoffSeriesCount = ReadSingleInt();

            Dictionary<int, int[]> runoffWeights = new Dictionary<int, int[]>();

            for (int h = 0; h < moduleHydrologyCount; h++)
            {
                int moduleNumber = ReadSingleInt();

                List<double> values = ReadMultipleDouble();
                if (ReadingFailed) { return false; }

                if ((values.Count <= 0 || values.Count > 10))
                {
                    ReadingFailed = true;
                    return false;
                }

                foreach(Module module in _modules.Where(m => m.Number == moduleNumber))
                {
                    module.AnnualRunoff = values[0];
                    for (int i = 0; i < values.Count - 1; i++)
                    {
                        if(!module.RunoffSeriesWeights.ContainsKey(i.ToString()))
                        {
                            module.RunoffSeriesWeights.Add(i.ToString(), values[i + 1]);
                            module.UseLocalInflow = true;
                        }
                    }
                }
            }

            for (int r = 0; r < runoffSeriesCount; r++)
            {
                int frad = ReadSingleInt();

                RunoffSeries runoffSeries = new RunoffSeries(ReadSingleString())
                {
                    Frad = frad
                };

                List<string> strings = ReadMultipleString(2);
                if (ReadingFailed) { return false; }

                runoffSeries.StationNumber = ReadSingleInt(strings[0]);
                runoffSeries.StationCode = strings[1];
                runoffSeries.ScalingFactor = ReadSingleDouble();
                ReadSingleString(); // File name

                hydrology.AddRunoffSeries(runoffSeries);
                foreach(Module module in _modules.Where(m => m.RunoffSeriesWeights.Keys.Contains(r.ToString())))
                {
                    module.RenameRunoffSeries(r.ToString(), runoffSeries.Name);
                }
            }

            if (ReadingFailed) { return false; }

            _project.Hydrology = hydrology;
            return true;
        }

        /// <summary>
        /// Sets job control of the project.
        /// Returns true when succesful.
        /// </summary>
        /// <returns></returns>
        private bool UnpackJobControl()
        {
            _lineIndex++;

            JobControl jobControl = new JobControl()
            {
                Name = ReadSingleString(),
                FirstYear = ReadSingleInt(),
                LastYear = ReadSingleInt(),
                FirmPowerLevel = ReadSingleInt(),
                ReservoirLevel = ReadSingleDouble(),

                SimulationType = ReadSingleInt(),
                FirstPeriod = ReadSingleInt(),
                LastPeriod = ReadSingleInt(),
            };

            List<int> outputs = ReadMultipleInt(4);
            if (ReadingFailed) { return false; }

            jobControl.OutputSystemDataSummary = outputs[0] == 1;
            jobControl.OutputWaterBalanceControl = outputs[1] == 1;
            jobControl.OutputPowerProductionStatistics = outputs[2] == 1;
            jobControl.OutputPowerCurtrailmentStatistics = outputs[3] == 1;

            jobControl.TextOutputLevel = ReadSingleInt();
            jobControl.ModuleOutput1 = ReadSingleInt();
            jobControl.ModuleOutput2 = ReadSingleInt();
            jobControl.ModuleOutput3 = ReadSingleInt();

            jobControl.OutputInBinary = ReadSingleInt() == 1;
            jobControl.BinaryFileName = ReadSingleString();

            if (ReadingFailed) { return false; }

            _project.JobControl = jobControl;
            return true;
        }
    }
}