﻿using NMag.Core.Utilization;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Text;
using System.Threading;

namespace NMag.Core
{
    // TODO stringbuilder
    // TODO use addcode for all cases

    public class DataSetWriter
    {
        private StringBuilder _stringBuilder;
        private Project _project;

        /// <summary>
        /// Writes project data into a data set.
        /// </summary>
        /// <param name="project"></param>
        public Result Write(Project project, string fileName)
        {
            if(File.Exists(fileName))
            {
                try
                {
                    File.Delete(fileName);
                }
                catch
                {
                    Thread.Sleep(10);
                    return new Result(false, string.Format("Can not delete old data set \"{0}\".", fileName));
                }
            }

            _project = project;
            _project.SortModules();
            _stringBuilder = new StringBuilder();

            WriteProjectInfo(true, "");
            WriteComments();

            WriteReservoirs();
            WritePowerPlants();
            WriteTransfers();
            WriteControlPoints();

            WriteEnergyMarket();
            WriteRestrictions();
            WriteOperationalStrategy();
            WriteHydrology();
            WriteJobControl();

            try
            {
                File.WriteAllText(fileName, _stringBuilder.ToString().Substring(0, _stringBuilder.Length - 2), Encoding.GetEncoding(1252));
            }
            catch
            {
                return new Result(false, string.Format("Can not write new data set \"{0}\".", fileName));
            }

            return new Result(true, "Data set successfully written.");
        }
        
        private void AddLine(Line line)
        {
            _stringBuilder.AppendLine(line.ToString());
        }

        private void AddSeparator()
        {
            _stringBuilder.AppendLine("***");
        }

        private void AddCurve(Curve curve, bool useGivenPrecision = false, int precisionX = -1, int precisionY = -1)
        {
            AddLine(new Line(curve.Points.Count));

            if(!useGivenPrecision)
            {
                precisionX = curve.Axes[0].DecimalValues ? -1 : 0;
                precisionY = curve.Axes[1].DecimalValues ? -1 : 0;
            }

            foreach (Point point in curve.Points)
            {
                AddLine(new Line(point.X, precisionX).AddValue(point.Y, precisionY));
            }
        }

        private void AddAnnualCurve(Curve curve)
        {
            List<double> values = new List<double>();

            for(int day = 0; day < 365; day++)
            {
                values.Add(curve.Points[day].Y);

                if(values.Count == 20)
                {
                    AddLine(new Line().AddValues(1, true, values.ToArray()));
                    values.Clear();
                }
            }

            AddLine(new Line().AddValues(1, true, values.ToArray()));
        }

        private void WriteProjectInfo(bool nms, string curFile)
        {
            if(nms)
            {
                AddLine(new Line(_project.ProjectSettings.TimeResolution).AddComment("* Number of timesteps in one year"));
            }
            else
            {
                AddLine(new Line(_project.ProjectSettings.TimeResolution).AddComment($"* Number of timesteps in one year.  File: {curFile}"));
            }

            AddLine(new Line(_project.Modules.Select(m => m.Number).Distinct().Count()).AddComment("* Total Number of Modules (*MA, *KR, *OF and *KO)"));

            AddLine(new Line(_project.Modules.Select(m => m.Units.Count(u => u.UnitType == UnitType.Reservoir)).Sum()).AddComment("* Number of reservoirs"));
            AddLine(new Line(_project.Modules.Select(m => m.Units.Count(u => u.UnitType == UnitType.PowerPlant)).Sum()).AddComment("* Number of power plants"));
            AddLine(new Line(_project.Modules.Select(m => m.Units.Count(u => u.UnitType == UnitType.Transfer)).Sum()).AddComment("* Number of transfer modules"));
            AddLine(new Line(_project.Modules.Select(m => m.Units.Count(u => u.UnitType == UnitType.ControlPoint)).Sum()).AddComment("* Number of control points"));

            AddLine(new Line(_project.ProjectSettings.MaxCurv).AddComment("* Max. permitted number of entries in a Table"));

            AddLine(new Line("*MO").AddComment("* Mode"));
            AddLine(new Line(_project.EnMagMode ? 1 : 0).AddComment("* 1=ENMAG mode, 0=nMAG mode"));
            AddSeparator();
        }

        private void WriteComments()
        {
            if(_project.ProjectSettings.Comments.Count > 0)
            {
                AddLine(new Line("*CO"));
                _project.ProjectSettings.Comments.ForEach(comment =>
                {
                    AddLine(new Line(comment));
                });
                AddSeparator();
            }
        }

        private void WriteReservoirs()
        {
            AddLine(new Line("*MA").AddComment("************ RESERVOIRS ************"));

            foreach (ReservoirUnit reservoir in UnitHelper.GetReservoirs(_project.Modules.SelectMany(m => m.Units).ToList()))
            {
                AddLine(new Line(reservoir.Module.Number).AddComment($"* Module #{reservoir.Module.Number}  -  {reservoir.Name}"));
                AddLine(new Line(reservoir.Name));

                AddLine(new Line(reservoir.LRWL, 1, true));
                AddLine(new Line(reservoir.HRWL, 1, true));
                AddLine(new Line(reservoir.Volume, 1, true));
                AddLine(new Line().AddValues(reservoir.Module.RoutingNumbers[0], reservoir.Module.RoutingNumbers[1], reservoir.Module.RoutingNumbers[2]));

                if(reservoir.UseVolumeCurve)
                {
                    AddLine(new Line(reservoir.VolumeCurve.Code));
                    AddCurve(reservoir.VolumeCurve);
                }

                if (reservoir.UseSpillFlowCurve)
                {
                    AddLine(new Line(reservoir.SpillFlowCurve.Code));
                    AddCurve(reservoir.SpillFlowCurve);
                }

                if (reservoir.UseGateDischargeCurve)
                {
                    AddLine(new Line(reservoir.GateDischargeCurve.Code));
                    AddLine(new Line(reservoir.OutletLevel, 1, true));
                    AddCurve(reservoir.GateDischargeCurve);
                }

                if(reservoir.UseEvaporationCurve)
                {
                    AddLine(new Line(reservoir.EvaporationCurve.Code));
                    AddCurve(reservoir.EvaporationCurve);
                }

                if (reservoir.Module.EnableRoutingCurve)
                {
                    AddLine(new Line(reservoir.Module.RoutingCurve.Code));
                    AddCurve(reservoir.Module.RoutingCurve);
                }
            }

            AddSeparator();
        }

        private void WritePowerPlants()
        {
            AddLine(new Line("*KR").AddComment("************ POWER PLANTS ************"));

            foreach(PowerPlantUnit powerPlant in UnitHelper.GetPowerPlants(_project.Modules.SelectMany(m => m.Units).ToList()))
            {
                AddLine(new Line(powerPlant.Module.Number).AddComment($"* Module #{powerPlant.Module.Number}  -  {powerPlant.Name}"));
                AddLine(new Line(powerPlant.Name));

                AddLine(new Line().AddValues(powerPlant.Module.RoutingNumbers[0], powerPlant.Module.RoutingNumbers[1], powerPlant.Module.RoutingNumbers[2]));
                AddLine(new Line(powerPlant.MaxDischarge, 1, true));
                AddLine(new Line(powerPlant.EnergyEquivalent, 1, true));

                if (powerPlant.UseOptionalData)
                {
                    AddLine(new Line(powerPlant.LoadEfficiencyCurve.Code));

                    AddLine(new Line(powerPlant.NominalHead, 1, true));
                    AddLine(new Line(powerPlant.IntakeLevel, 1, true));
                    AddLine(new Line(powerPlant.TailWaterLevel, 1, true));
                    AddLine(new Line(powerPlant.HeadLossCoefficient, 1, true));

                    AddCurve(powerPlant.LoadEfficiencyCurve);
                }

                if (powerPlant.UsePeakingCurve)
                {
                    AddLine(new Line(powerPlant.PeakingCurve.Code));
                    AddCurve(powerPlant.PeakingCurve);
                }

                if (powerPlant.Module.EnableRoutingCurve)
                {
                    AddLine(new Line(powerPlant.Module.RoutingCurve.Code));
                    AddCurve(powerPlant.Module.RoutingCurve);
                }
            }

            AddSeparator();
        }

        private void WriteTransfers()
        {
            AddLine(new Line("*OF").AddComment("************ TRANSFER MODULES ************"));

            foreach (TransferUnit transfer in UnitHelper.GetTransfers(_project.Modules.SelectMany(m => m.Units).ToList()))
            {
                AddLine(new Line(transfer.Module.Number).AddComment($"* Module #{transfer.Module.Number}  -  {transfer.Name}"));
                AddLine(new Line(transfer.Name));

                AddLine(new Line().AddValues(transfer.Module.RoutingNumbers[0], transfer.Module.RoutingNumbers[1], transfer.Module.RoutingNumbers[2]));

                switch(transfer.Mode)
                {
                    case TransferUnit.TransferMode.FixedCapacity:
                        AddLine(new Line((int)transfer.Mode));
                        AddLine(new Line(transfer.Capacity, 1, true));
                        break;

                    case TransferUnit.TransferMode.VariableCapacity:
                        AddLine(new Line((int)transfer.Mode));
                        AddLine(new Line().AddValues(1, true, transfer.InletElevation, transfer.OutletElevation, transfer.HeadLossCoefficient));
                        break;

                    case TransferUnit.TransferMode.HQCapacityCurve:
                        AddLine(new Line((int)transfer.Mode));
                        AddLine(new Line(transfer.IntakeElevation, 1, true));
                        AddLine(new Line(transfer.TailwaterElevation, 1, true));
                        AddCurve(transfer.HQCapacityCurve);
                        break;

                    default:
                        break;
                }

                if(transfer.Module.EnableRoutingCurve)
                {
                    AddLine(new Line(transfer.Module.RoutingCurve.Code));
                    AddCurve(transfer.Module.RoutingCurve);
                }
            }

            AddSeparator();
        }

        private void WriteControlPoints()
        {
            AddLine(new Line("*KO").AddComment("************ CONTROL POINTS ************"));

            foreach (ControlPointUnit controlPoint in UnitHelper.GetControlPoints(_project.Modules.SelectMany(m => m.Units).ToList()))
            {
                AddLine(new Line(controlPoint.Module.Number).AddComment($"* Module #{controlPoint.Module.Number}  -  {controlPoint.Name}"));
                AddLine(new Line(controlPoint.Name));

                AddLine(new Line().AddValues(controlPoint.Module.RoutingNumbers[0], controlPoint.Module.RoutingNumbers[1], controlPoint.Module.RoutingNumbers[2]));
                
                if(controlPoint.Module.EnableRoutingCurve)
                {
                    AddLine(new Line(controlPoint.Module.RoutingCurve.Code));
                    AddCurve(controlPoint.Module.RoutingCurve);
                }
            }

            AddSeparator();
        }

        private void WriteEnergyMarket()
        {
            AddLine(new Line("*KM").AddComment("************ POWER MARKET ************"));

            EnergyMarket energyMarket = _project.EnergyMarket;

            AddLine(new Line(energyMarket.FirmPowerLevel, 1, true));
            AddLine(new Line(energyMarket.FirmPowerPrice, 1, true));
            AddLine(new Line().AddValues(energyMarket.SummerStart, energyMarket.SummerEnd));
            AddLine(new Line((int)energyMarket.Mode));
            
            switch(energyMarket.Mode)
            {
                // No curve code needed.
                case EnergyMarket.DistributionMode.DailyLoad:
                    AddAnnualCurve(energyMarket.DailyLoadCurve);
                    break;

                // No curve code needed.
                case EnergyMarket.DistributionMode.Vairable:
                    AddCurve(energyMarket.VariableLoadCurve);
                    break;

                default:
                    break;
            }

            AddLine(new Line(energyMarket.PreferenceFunctions.Count).AddComment("* Number of preference functions"));

            for (int i = 0; i < energyMarket.PreferenceFunctions.Count; i++)
            {
                PreferenceFunction preferenceFunction = energyMarket.PreferenceFunctions[i].Clone();
                preferenceFunction.PreferenceCurve.ConvertToDiscrete();

                AddLine(new Line(i + 1));
                AddLine(new Line().AddValues(preferenceFunction.Start, preferenceFunction.End));
                AddCurve(preferenceFunction.PreferenceCurve);       // No curve code needed
                AddLine(new Line().AddValues(preferenceFunction.RationingStart, preferenceFunction.SellingStart));
            }

            AddSeparator();
        }

        private void WriteRestrictions()
        {
            AddLine(new Line("*RS").AddComment("************ RESTRICTIONS ************"));

            foreach (Module module in _project.Modules.Where(m => m.UsesRestrictions))
            {
                string header = $"* Module #{module.Number}  -  {module.Name}";
                
                if(module.UseMinimumFlow)
                {
                    AddLine(new Line(module.MinimumFlowCurve.Code));
                    AddLine(new Line(module.Number).AddComment(header));
                    AddCurve(module.MinimumFlowCurve, true, 0, 3);
                }

                if (module.UseBypassRelease)
                {
                    AddLine(new Line(module.BypassReleaseCurve.Code));
                    AddLine(new Line(module.Number).AddComment(header));
                    AddCurve(module.BypassReleaseCurve, true, 0, 3);
                }

                if (module.UseMinimumLevel)
                {
                    AddLine(new Line(module.MinimumLevelCurve.Code));
                    AddLine(new Line(module.Number).AddComment(header));
                    AddCurve(module.MinimumLevelCurve, true, 0, 3);
                }

                if (module.UseMaximumLevel)
                {
                    AddLine(new Line(module.MaximumLevelCurve.Code));
                    AddLine(new Line(module.Number).AddComment(header));
                    AddCurve(module.MaximumLevelCurve, true, 0, 3);
                }
            }

            AddSeparator();
        }

        private void WriteOperationalStrategy()
        {
            AddLine(new Line("*ST").AddComment("************ OPERATIONAL STRATEGY ************"));

            if (_project.EnMagMode)
            {
                WriteEnMagOperationalStrategy();
            }
            else
            {
                WriteNMagOperationalStrategy();
            }

            AddSeparator();
        }

        private void WriteNMagOperationalStrategy()
        {
            foreach (Module module in _project.Modules.Where(m => m.UseOperationalStrategy && m.NMagOperationalStrategy != null))
            {
                string header = $"* Module #{module.Number}  -  {module.Name}";

                switch (module.NMagOperationalStrategy.Strategy)
                {
                    case NMagOperationalStrategy.OperationalStrategy.Automatic:
                        AddLine(new Line((int)module.NMagOperationalStrategy.Strategy));
                        AddLine(new Line(module.Number).AddComment(header));
                        break;

                    case NMagOperationalStrategy.OperationalStrategy.ReservoirRelease:
                        AddLine(new Line((int)module.NMagOperationalStrategy.Strategy));
                        AddLine(new Line(module.Number).AddValue(module.NMagOperationalStrategy.AnnualRelease,1, true).AddComment(header));
                        AddCurve(module.NMagOperationalStrategy.ReservoirReleaseCurve);
                        break;

                    case NMagOperationalStrategy.OperationalStrategy.ReservoirGuide:
                        AddLine(new Line((int)module.NMagOperationalStrategy.Strategy));
                        AddLine(new Line(module.Number).AddComment(header));
                        AddCurve(module.NMagOperationalStrategy.ReservoirGuideCurve);
                        break;

                    default:
                        break;
                }
            }
        }

        private void WriteEnMagOperationalStrategy()
        {
            if(!_project.EnMagMode) { return; }

            Module module = _project.Modules
                .FirstOrDefault(m => m.UseOperationalStrategy && m.EnMagOperationalStrategy != null);

            if (module == null) { return; }

            string header = $"* Module #{module.Number}  -  {module.Name}";

            switch (module.EnMagOperationalStrategy.Strategy)
            {
                case EnMagOperationalStrategy.OperationalStrategy.Unconditional:
                    AddLine(new Line((int)module.EnMagOperationalStrategy.Strategy));
                    AddLine(new Line(module.Number).AddComment(header));
                    break;

                case EnMagOperationalStrategy.OperationalStrategy.GuideCurve:
                    AddLine(new Line((int)module.EnMagOperationalStrategy.Strategy));
                    AddLine(new Line(module.Number).AddComment(header));
                    AddCurve(module.EnMagOperationalStrategy.GuideCurve);
                    break;

                case EnMagOperationalStrategy.OperationalStrategy.RuleCurve:
                    AddLine(new Line((int)module.EnMagOperationalStrategy.Strategy));
                    AddLine(new Line(module.Number).AddComment(header));
                    AddCurve(module.EnMagOperationalStrategy.RationingRuleCurve);
                    AddCurve(module.EnMagOperationalStrategy.SurplusRuleCurve);
                    break;

                default:
                    break;
            }
        }

        private void WriteHydrology()
        {
            AddLine(new Line("*HY").AddComment("************ HYDROLOGICAL INPUT DATA ************"));

            Hydrology hydrology = _project.Hydrology;

            int firstYear = hydrology.RunoffSeries.Select(s => s.HydroCurve.FirstYear).Max();
            int lastYear = hydrology.RunoffSeries.Select(s => s.HydroCurve.LastYear).Min();

            AddLine(new Line().AddValues(firstYear, lastYear));
            AddLine(new Line(_project.Modules.Count(m => m.UseLocalInflow)));
            AddLine(new Line(hydrology.RunoffSeries.Count));

            foreach (Module module in _project.Modules.Where(m => m.UseLocalInflow))
            {
                AddLine(new Line(module.Number).AddComment($"* Module #{module.Number}  -  {module.Name}"));

                Line line = new Line($"   {module.AnnualRunoff.Format(1, true)},");
                
                foreach (RunoffSeries runoffSeries in hydrology.RunoffSeries)
                {
                    line.AddValue(module.RunoffSeriesWeights[runoffSeries.Name], 1, true);
                }

                AddLine(line);
            }

            foreach (RunoffSeries runoffSeries in hydrology.RunoffSeries)
            {
                AddLine(new Line(runoffSeries.Frad));
                AddLine(new Line(runoffSeries.Name));
                AddLine(new Line(runoffSeries.StationNumber).AddText(runoffSeries.StationCode, true));
                AddLine(new Line(runoffSeries.ScalingFactor, 1, true));
                AddLine(new Line(runoffSeries.Name.Replace(" ", "") + ".vsf"));
            }

            AddSeparator();
        }

        private void WriteJobControl()
        {
            AddLine(new Line("*JK").AddComment("************ JOB CONTROL ************"));

            JobControl jobControl = _project.JobControl;

            AddLine(new Line(jobControl.Name));

            AddLine(new Line(jobControl.FirstYear));
            AddLine(new Line(jobControl.LastYear));
            AddLine(new Line(jobControl.FirmPowerLevel));
            AddLine(new Line(jobControl.ReservoirLevel, 1, true));

            AddLine(new Line(jobControl.SimulationType));
            AddLine(new Line(jobControl.FirstPeriod));
            AddLine(new Line(jobControl.LastPeriod));
            AddLine(new Line(jobControl.GetOutputControlString()));

            AddLine(new Line(jobControl.TextOutputLevel));
            AddLine(new Line(jobControl.ModuleOutput1));
            AddLine(new Line(jobControl.ModuleOutput2));
            AddLine(new Line(jobControl.ModuleOutput3));
            AddLine(new Line(jobControl.OutputInBinary ? 1 : 0));
            AddLine(new Line(jobControl.BinaryFileName == "" ? "None" : jobControl.BinaryFileName));

            foreach(TimeSeries timeSeries in jobControl.TimeSeries)
            {
                AddLine(new Line(timeSeries.ToString()));
            }

            AddSeparator();
            AddLine(new Line("*EX"));
        }

        private class Line
        {
            private const int COMMENT_START = 30;

            private string _line = "";

            public Line() { }

            public Line(int value)
            {
                AddValue(value);
            }

            public Line(double value, int precision, bool useMinimum = false)
            {
                AddValue(value, precision, useMinimum);
            }

            public Line(string text, bool addComma = false)
            {
                AddText(text, addComma);
            }

            public override string ToString()
            {
                return _line;
            }

            public Line AddValue(int value)
            {
                _line += value.ToString() + ",";
                return this;
            }

            public Line AddValues(params int[] values)
            {
                _line += string.Join("", values.Select(v => v.ToString() + ","));
                return this;
            }

            public Line AddValue(double value, int precision, bool useMinimum = false)
            {
                _line += value.Format(precision, useMinimum) + ",";
                return this;
            }

            public Line AddValues(int precision, bool useMinimum = false, params double[] values)
            {
                _line += string.Join("", values.Select(v => v.Format(precision, useMinimum) + ","));
                return this;
            }

            public Line AddText(string text, bool addComma = false)
            {
                _line += text;
                if (addComma) { _line += ","; }
                return this;
            }

            public Line AddTexts(bool addComma = false, params string[] text)
            {
                _line += string.Join("", text);
                if (addComma) { _line += ","; }
                return this;
            }

            public Line AddComment(string comment, int commentStart = COMMENT_START)
            {
                if(_line.Length <= commentStart)
                {
                    _line = _line.AddTail(' ', commentStart - _line.Length) + comment;
                }

                return this;
            }
        }
    }
}