﻿using NMag.Core.Utilization;
using System;
using System.Collections.Generic;
using System.Data;
using System.IO;
using System.Linq;
using System.Windows.Forms;
using Excel = Microsoft.Office.Interop.Excel;

namespace NMag.Core
{
    public class DataReader
    {
        private Dictionary<int, int> _firstYears = new Dictionary<int, int>();

        private string _dataSource = "";

        public DataTable Data { get; private set; }
        public int FirstYear { get; private set; }

        public Result ReadText(string path)
        {
            _dataSource = path;

            List<DataLine> lines = new List<DataLine>();

            try
            {
                lines = SplitLines(File.ReadAllLines(path).ToList());
            }
            catch(Exception exception)
            {
                return new Result(false, $"Could not read file {path}.\nException: {exception}");
            }

            return ReadLines(lines);
        }

        public Result ReadExcel(string path)
        {
            _dataSource = path;

            if (Type.GetTypeFromProgID("Excel.Application") == null)
            {
                return new Result(false, "Cannot open file in Excel.\nMake sure Excel is installed on this pc.");
            }

            Excel.Application xlApp = null;
            Excel.Workbook workBook = null;

            List<string> data = new List<string>();

            try
            {
                xlApp = new Excel.Application()
                {
                    DisplayAlerts = false,
                    Visible = true
                };

                workBook = xlApp.Workbooks.Open(path, 0, false, 5, "", "", true, Excel.XlPlatform.xlWindows, "\t", true, false, 0, true, 1, 0);

                Excel.Worksheet worksheet = workBook.Sheets[1];

                worksheet.UsedRange.Replace("+ACI-", " ");
                worksheet.UsedRange.Replace("\r", " ");
                worksheet.UsedRange.Replace("\n", ", ");
                worksheet.UsedRange.Replace(";", ", ");

                worksheet.UsedRange.Copy();

                data = Clipboard.GetText()
                    .Split(new string[] { "\r\n" }, StringSplitOptions.None)
                    .ToList();

                worksheet.Cells[1, 1].Copy();
            }
            catch (Exception e)
            {
                return new Result(false, $"An error occured when reading the Excel file.\nException: \"{e.Message}\"");
            }
            finally
            {
                if (workBook != null)
                {
                    workBook.Close(SaveChanges: false);
                }

                xlApp.Quit();
            }

            return ReadLines(SplitLines(data));
        }

        public bool GetData(string xColumn, string yColumn, out List<Point> points, out string dataSource)
        {
            if(Data == null ||
                !Data.Columns.Contains(xColumn ?? "") ||
                !Data.Columns.Contains(yColumn ?? ""))
            {
                dataSource = "";
                points = null;
                return false;
            }

            List<double>[] data = new List<double>[]
            {
                GetColumn(Data.Columns.IndexOf(xColumn)),
                GetColumn(Data.Columns.IndexOf(yColumn))
            };

            points = new List<Point>();

            if(data.Contains(null) || data[0].Count != data[1].Count)
            {
                dataSource = "";
                return false;
            }

            for (int r = 0; r < data.First().Count; r++)
            {
                points.Add(new Point(data[0][r], data[1][r]));
            }

            dataSource = _dataSource;
            return true;
        }

        private Result ReadLines(List<DataLine> lines)
        {
            if(lines.Count == 0)
            {
                return new Result(false, "No data to be read.");
            }

            Data = new DataTable();
            FirstYear = 0;
            _firstYears.Clear();

            int columnCount = lines.First().Values.Count;

            CreateColumns(ref lines, columnCount);

            CreateRows(lines, columnCount);

            return new Result(true);
        }

        private void RemoveLeapDays(int columnIndex)
        {
            if (columnIndex < 0 || columnIndex >= Data.Columns.Count)
            {
                return;
            }

            for (int r = Data.Rows.Count - 1; r >= 0; r--)
            {
                string text = Data.Rows[r].ItemArray[columnIndex].ToString();

                if(DateHelper.ParseDate(text, true, out DateTime date) && date.Month == 2 && date.Day == 29)
                {
                    Data.Rows.RemoveAt(r);
                }
            }
        }

        private List<double> GetColumn(int columnIndex)
        {
            if (columnIndex < 0 || columnIndex >= Data.Columns.Count)
            {
                return null;
            }

            if (_firstYears.ContainsKey(columnIndex))
            {
                FirstYear = _firstYears[columnIndex];
                RemoveLeapDays(columnIndex);
            }

            List<double> column = new List<double>();

            foreach (DataRow row in Data.Rows)
            {
                string text = row.ItemArray[columnIndex].ToString();

                if(_firstYears.ContainsKey(columnIndex))
                {
                    DateHelper.ParseDate(text, true, out DateTime date);

                    column.Add(DateHelper.DayOfYear(_firstYears[columnIndex], date));
                }
                else
                {
                    double.TryParse(text, out double value);
                    column.Add(value);
                }
            }

            return column;
        }

        private void CreateColumns(ref List<DataLine> lines, int columnCount)
        {
            List<string> columnNames = GetColumnNames(lines.First());

            if (columnNames.Count == columnCount)
            {
                columnNames.ForEach(h => CreateColumn(h));

                lines = lines.Skip(1).ToList();
            }
            else
            {
                for (int h = 0; h < columnCount; h++)
                {
                    CreateColumn("");
                }
            }
        }

        private void CreateColumn(string columnName)
        {
            if(columnName == "")
            {
                Data.Columns.Add(new DataColumn() { ReadOnly = true });
            }
            else
            {
                Data.Columns.Add(new DataColumn(columnName) { ReadOnly = true });
            }
        }

        private List<string> GetColumnNames(DataLine line)
        {
            List<string> names = new List<string>();
            bool foundString = false;

            foreach (string text in line.Values)
            {
                if (!ParseDouble(text, out double value) &&
                    !DateHelper.ParseDate(text, true, out DateTime date) &&
                    !DateHelper.ParseDate(text, false, out date))
                {
                    foundString = true;
                }

                CreateColumnName(text, names);
            }

            if (!foundString)
            {
                return new List<string>();
            }

            return names;
        }

        private void CreateColumnName(string text, List<string> columnNames)
        {
            string name = text
                .LettersAndDigitsOnly()
                .Replace(" ", "_");

            columnNames.Add(StringHelper.GetUniqueName(name, columnNames));
        }

        private List<DateTime> GetDates(List<string> values)
        {
            bool dayLast = false;

            if (values.Count > 1)
            {
                dayLast = values[1].Last() - values[0].Last() == 1;
            }

            List<DateTime> dates = new List<DateTime>();

            foreach (string value in values)
            {
                if(DateHelper.ParseDate(value, dayLast, out DateTime date))
                {
                    dates.Add(date);
                }
            }

            return (dates.Count == values.Count) ? dates : new List<DateTime>();
        }

        private void FormatData(List<DataLine> lines, int columnCount)
        {
            for (int c = 0; c < columnCount; c++)
            {
                List<string> values = lines
                    .Select(l => l.Values[c])
                    .ToList();

                List<DateTime> dates = GetDates(values);

                if (dates.Count == values.Count)
                {
                    _firstYears.Add(c, dates.First().Year);

                    for (int d = 0; d < dates.Count; d++)
                    {
                        lines[d].Values[c] = dates[d].ToString("yyyy-MM-dd");
                    }
                }
                else
                {
                    for (int v = 0; v < values.Count; v++)
                    {
                        if (!ParseDouble(lines[v].Values[c], out double value))
                        {
                            lines[v].Values[c] = "INVALID";
                            continue;
                        }

                        lines[v].Values[c] = value.Format(3);
                    }
                }
            }
        }

        private void CreateRows(List<DataLine> lines, int columnCount)
        {
            FormatData(lines, columnCount);

            foreach(DataLine line in lines)
            {
                DataRow row = Data.NewRow();
                row.ItemArray = line.Values.ToArray();
                Data.Rows.Add(row);
            }
        }

        private bool ParseDouble(string text, out double value)
        {
            if (!UsesDotForDecimals(text))
            {
                text = text.Replace('.', '@').Replace(',', '.').Replace('@', ',');
            }

            return double.TryParse(text, out value);
        }

        private bool UsesDotForDecimals(string text)
        {
            return (text.IndexOf(',') > -1) ?
                ((text.IndexOf('.') > text.IndexOf(',')) ? true : false) :
                true;
        }

        private List<DataLine> SplitLines(List<string> data)
        {
            List<DataLine> lines = new List<DataLine>();

            if(data.First().Count(c => c == '\t') > 0)
            {
                lines = SplitDelimitedData(data, '\t');
            }
            else if(data.First().Count(c => c == ';') > 0)
            {
                lines = SplitDelimitedData(data, ';');
            }
            else
            {
                lines = SplitDelimitedData(data, ',');
            }
            
            int columnCount = lines.Select(l => l.Values.Count).Max();

            foreach(DataLine line in lines.Where(l => l.Values.Count < columnCount))
            {
                for (int v = line.Values.Count; v < columnCount; v++)
                {
                    line.Values.Add("0");
                }
            }
            
            return lines;
        }

        private List<DataLine> SplitDelimitedData(List<string> data, char delimiter)
        {
            List<DataLine> lines = new List<DataLine>();

            data.Where(l => l != "")
                .ToList()
                .ForEach(l => lines.Add(new DataLine(l.Split(delimiter).ToList())));

            return lines;
        }
    }
}