﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace NMag.Core.Optimization
{
    public class CandidateSolution
    {
        public CandidateSolution(Generation generation, int number, List<Chromosome> chromosomes)
        {
            Generation = generation;
            Chromosomes = chromosomes;
            Number = number;
        }

        public CandidateSolution(Generation generation, int number, CandidateSolution parent)
        {
            Generation = generation;
            Parent = parent;
            Number = number;
            Chromosomes = parent.Chromosomes;
        }

        public Generation Generation { get; private set; }
        public int Number { get; private set; }

        public CandidateSolution Parent { get; private set; }
        public List<Chromosome> Chromosomes { get; private set; }

        public double FitNess { get; set; }

        public bool IsMutated { get; private set; }
        public bool IsDead { get; private set; }

        public void Mutate(int propability, double speed, Point range, Random mutator)
        {
            List<Chromosome> newGenoType = new List<Chromosome>();

            foreach(Chromosome chromosome in Chromosomes)
            {
                if(mutator.Next(1, 100) <= propability)
                {
                    IsMutated = true;
                    newGenoType.Add(MutateChromosome(chromosome, speed, range, mutator));
                }
                else { newGenoType.Add(chromosome); }
            }

            Chromosomes = newGenoType;
        }

        public void Evaluate(Action<CandidateSolution> evaluation)
        {
            List<ReservoirGuideCurve> dna = Chromosomes.Select(c => new ReservoirGuideCurve(c.BasePairs)).ToList();

            if(dna.Count(c => !c.Validate().Succes) > 0 || HasSpikes())
            {
                IsDead = true;
                return;
            }

            evaluation(this);
        }
       
        private bool HasSpikes()
        {
            foreach (Chromosome chromosome in Chromosomes)
            {
                for (int b = 1; b < chromosome.BasePairs.Count; b++)
                {
                    Point basePair1 = chromosome.BasePairs[b - 1];
                    Point basePair2 = chromosome.BasePairs[b];

                    if (Math.Abs((basePair2.Y - basePair1.Y) / (basePair2.X - basePair1.X)) > 2)
                    {
                        return true;
                    }
                }
            }

            return false;
        }

        private Chromosome MutateChromosome(Chromosome chromosome, double speed, Point range, Random mutator)
        {
            Chromosome mutatedChromosome = chromosome.Clone();

            for (int c = 0; c < chromosome.BasePairs.Count * speed / 100; c++)
            {
                int mutationIndex = mutator.Next(0, chromosome.BasePairs.Count);

                Point basePair = mutatedChromosome.BasePairs[mutationIndex];

                if (mutationIndex == 0 || mutationIndex == chromosome.BasePairs.Count - 1)
                {
                    mutatedChromosome.BasePairs[mutationIndex] = new Point(basePair.X, basePair.Y + Mutation((int)range.Values[0]));
                }
                else
                {
                    mutatedChromosome.BasePairs[mutationIndex] = new Point(basePair.X + Mutation((int)range.Values[0]), basePair.Y + Mutation((int)range.Values[0]));
                }
            }

            return mutatedChromosome;
        }

        private double Mutation(int range)
        {
            return Generation.Mutator.Next(1, range * 100) / 50.0 - range / 2.0;
        }
    }
}
