﻿using NMag.Core.Simulation;
using NMag.Core.Utilization;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Threading;

namespace NMag.Core.Optimization
{
    public class Optimizer
    {
        private Action<string> _updateStatusAction;

        private Project _project;
        private Thread _simulationThread;

        private Simulation.Simulation _simulation;

        public Optimizer(Action<string> updateStatusAction, string simulationName)
        {
            _updateStatusAction = updateStatusAction;
            _simulation = new Simulation.Simulation(simulationName);
        }

        public bool IsRunning
        {
            get { return _simulationThread.IsAlive; }
        }

        public void Start(Project project, PopulationControl populationControl)
        {
            if (project == null || !project.Validate().Succes) { return; }

            _project = project.Clone();

            _simulationThread = new Thread(() => Run(populationControl)) { IsBackground = true };
            _simulationThread.Start();
        }

        public void Abort()
        {
            _simulationThread.Abort();
            Report("Optimizer Aborted");
        }

        private List<Module> GetReservoirModules(List<Module> modules)
        {
            return modules.Where(m => m.ModuleType == ModuleType.Reservoir || m.ModuleType == ModuleType.Combined).ToList();
        }

        private SimulationResult SimulateAutomaticBalancing(Project project, Simulation.Simulation simulation)
        {
            project.EnMagMode = false;
            foreach (Module reservoir in GetReservoirModules(project.Modules))
            {
                reservoir.NMagOperationalStrategy = new NMagOperationalStrategy()
                {
                    Strategy = NMagOperationalStrategy.OperationalStrategy.Automatic
                };
            }
        
            return simulation.Run(project, ClearLevel.All);
        }

        private SimulationResult SimulateUnconditionalProduction(Project project, Simulation.Simulation simulation)
        {
            project.EnMagMode = true;
            GetReservoirModules(project.Modules).First().EnMagOperationalStrategy = new EnMagOperationalStrategy()
            { Strategy = EnMagOperationalStrategy.OperationalStrategy.Unconditional };

            return simulation.Run(project, ClearLevel.All);
        }

        private SimulationResult SimulateGuidedReservoir(Project project, Simulation.Simulation simulation, List<ReservoirGuideCurve> guideCurves)
        {
            project.EnMagMode = false;
            for (int r = 0; r < guideCurves.Count; r++)
            {
                GetReservoirModules(project.Modules)[r].NMagOperationalStrategy = new NMagOperationalStrategy()
                {
                    Strategy = NMagOperationalStrategy.OperationalStrategy.ReservoirGuide,
                    ReservoirGuideCurve = guideCurves[r]
                };
            }


            return simulation.Run(project, ClearLevel.Minimum);
        }

        private IEnumerable<ReservoirGuideCurve> CreateGuideCurves(List<Chromosome> chromosomes)
        {
            foreach (Chromosome chromosome in chromosomes)
            {
                yield return new ReservoirGuideCurve(chromosome.BasePairs.Select(b => b.Clone()).ToList());
            }
        }

        private IEnumerable<ReservoirGuideCurve> CreateGuideCurves(List<StorageCurve> storageCurves, List<ReservoirUnit> reservoirs, int populationSize)
        {
            for (int r = 0; r < reservoirs.Count; r++)
            {
                List<double> storage = storageCurves[r].Values[1];
                double totalVolume = reservoirs[r].Volume;

                List<Point> guidePoints = new List<Point> { new Point(1, storage[0] / totalVolume * 100) };
                for (int i = 1; i < populationSize; i++)
                {
                    int day = (int)Math.Min(365, Math.Max(1, (365.0 / populationSize * i)));
                    double value = Math.Min(100, Math.Max(0, storage[day - 1] / totalVolume * 100));

                    guidePoints.Add(new Point(day, value));
                }
                guidePoints.Add(new Point(365, storage[364] / totalVolume * 100));

                yield return new ReservoirGuideCurve(guidePoints);
            }
        }

        private IEnumerable<StorageCurve> GetStorateCurves(int firstYear, List<ReservoirUnit> reservoirs, SimulationResult results)
        {
            foreach (ReservoirUnit reservoir in reservoirs)
            {
                List<Point> points = new List<Point>();
                for (int d = 0; d < results.Periods.Count; d++)
                {
                    points.Add(new Point(d, Math.Round(results.ModuleResults.First(m => m.Module.Number == reservoir.Module.Number).Storage[d], 3)));
                }

                yield return new StorageCurve(points, firstYear).GetAverage();
            }
        }

        private void PrintCurve(Curve curve, string path)
        {
            new DataWriter().Write(Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Simulation), path), curve, ",");
        }

        private List<Point> GetAverage(List<Point> points)
        {
            return new HydroCurve(points, 0).GetAverage().Points;
        }

        private void Run(PopulationControl populationControl)
        {
            _project.SortModules();

            List<ReservoirUnit> reservoirs = _project.Modules
                .Where(m => m.ModuleType == ModuleType.Reservoir || m.ModuleType == ModuleType.Combined)
                .Select(m => UnitHelper.GetReservoirs(m.Units).First())
                .ToList();

            if (reservoirs.Count == 0)
            {
                Report("Could not find a reservoir.");
                return;
            }

            SimulationResult results;

            if (reservoirs.Count == 1)
            {
                results = SimulateUnconditionalProduction(_project, _simulation);

                if (results == null)
                {
                    Report("Could not run simulation.");
                    return;
                }

                Report($"Unconditional Run - Net Benefit: {results.NetBenefit.Format(3)}");
                new ResultWriter().Write(results, _project.Name, _simulation.Name, "Unconditional simulation");
            }
            else
            {
                results = SimulateAutomaticBalancing(_project, _simulation);

                if (results == null)
                {
                    Report("Could not run simulation.");
                    return;
                }

                Report($"Auto Balancing Run - Net Benefit: {results.NetBenefit.Format(3)}");
                new ResultWriter().Write(results, _project.Name, _simulation.Name, "Auto balance simulation");
            }

            string simulationFolder = Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Project, _project.Name), _simulation.Name);
            populationControl.UpdateStatusAction = _updateStatusAction;
            Report($"Opimization Started " +
                $"G{populationControl.GenerationCount} " +
                $"N{populationControl.PopulationSize} " +
                $"C{populationControl.GeneComplexity} " +
                $"P{populationControl.MutationProbability} " +
                $"S{populationControl.MutationSpeed} " +
                $"R{populationControl.MutationRange.X};");
            
            List<StorageCurve> storageCurves = GetStorateCurves(_project.JobControl.FirstYear, reservoirs, results).ToList();
            new DataWriter().Write(Path.Combine(simulationFolder, "first storage curve.csv"), storageCurves.First(), ",");

            List<ReservoirGuideCurve> guideCurves = CreateGuideCurves(storageCurves, reservoirs, populationControl.GeneComplexity).ToList();
            new DataWriter().Write(Path.Combine(simulationFolder, "first guide curve.csv"), guideCurves.First(), ",");

            results = SimulateGuidedReservoir(_project, _simulation, guideCurves);
            new ResultWriter().Write(results, _project.Name, _simulation.Name, "first guided simulation");


            List<Chromosome> chromosomes = guideCurves.Select(c => new Chromosome(c.Points.Select(p => p.Clone()).ToList())).ToList();

            populationControl.Evolve(chromosomes, EvaluateGuideCurve);

            CandidateSolution bestSolution = populationControl.Fittest;
            if(bestSolution == null)
            {
                Report("Could not find a good solution.");
                return;
            }

            new ResultWriter().Write(SimulateGuidedReservoir(_project, _simulation, CreateGuideCurves(bestSolution.Chromosomes).ToList()), _project.Name, _simulation.Name, "Best fit curve simulation");
            populationControl.PrintEvolution(Path.Combine(simulationFolder, $"Evolution {DateTime.Now.ToString("HHmmss")}.csv"));

            Report("Opimization Ended;");
            return;
        }

        private void EvaluateGuideCurve(CandidateSolution solution)
        {
            SimulationResult results = SimulateGuidedReservoir(_project, _simulation, CreateGuideCurves(solution.Chromosomes).ToList());

            if (results == null)
            {
                Report($"Generation {solution.Generation.Number} #{solution.Number} - Simulator could not run.");
                solution.FitNess = -99999999999999999;
            }
            else
            {
                Report($"Generation {solution.Generation.Number} #{solution.Number} - Net Benefit: {results.NetBenefit.Format(3)}");
                solution.FitNess = results.NetBenefit;
            }
        }

        private void Report(string message)
        {
            _updateStatusAction(message);
        }
    }
}