﻿using NMag.Core.Utilization;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;

namespace NMag.Core.Optimization
{
    public class PopulationControl
    {
        public PopulationControl()
        {

        }

        public PopulationControl(int generationCount, int populationSize, int geneComplexity , int mutationProbability, int mutationSpeed, Point mutationRange)
        {
            GenerationCount = generationCount;
            PopulationSize = populationSize;
            GeneComplexity = geneComplexity;

            MutationProbability = mutationProbability;
            MutationSpeed = mutationSpeed;
            MutationRange = mutationRange;
        }

        public Action<string> UpdateStatusAction { get; set; }

        public int GenerationCount { get; set; } = 50;

        public int PopulationSize { get; set; } = 50;
        public int GeneComplexity { get; set; } = 30;

        public int MutationProbability { get; set; } = 15;
        public int MutationSpeed { get; set; } = 10;
        public Point MutationRange { get; set; } = new Point(2, 2);

        public TimeSpan Age { get; private set; } = new TimeSpan();
        public List<CandidateSolution> Evolution { get; private set; } = new List<CandidateSolution>();
        public CandidateSolution Fittest
        {
            get
            {
                if(Evolution.Count == 0) { return null; }
                double maxFitness = Evolution.Select(s => s.FitNess).Max();
                return Evolution.First(s => s.FitNess == maxFitness);
            }
        }

        public Generation Generation { get; private set; }

        public void Evolve(List<Chromosome> seeds, Action<CandidateSolution> evaluation)
        {
            DateTime evolutionStart = DateTime.Now;
            Evolution = new List<CandidateSolution>();
            Generation = new Generation(this, evaluation);

            Generation.Populate(seeds);
            
            double maxFitness = 0;

            for (int i = 0; i < GenerationCount; i++)
            {
                Generation.EvaluateGeneration();
                CandidateSolution fittest = Generation.Fittest;
                if(fittest == null)
                {
                    Age = DateTime.Now - evolutionStart;
                    return;
                }

                Evolution.Add(fittest);
                maxFitness = i == 0 ? fittest.FitNess : Math.Max(maxFitness, fittest.FitNess);

                UpdateStatusAction($"Current fittest: {fittest.FitNess.Format(3)}");
                Generation = Generation.GenerateOffspring();
            }

            Age = DateTime.Now - evolutionStart;
        }

        // TODO transpose function
        public void PrintEvolution(string fileName)
        {
            List<DataLine> dailyValues = new List<DataLine>()
            {
                new DataLine(new List<string>()
                {
                    "Evolution time: " + Age.ToString(@"hh\:mm\:ss"),
                    $"Generations: {GenerationCount}",
                    $"Population size: {PopulationSize}",
                    $"Gene complexity: {GeneComplexity}",
                    $"Mutation Probability: {MutationProbability}",
                    $"Mutation Speed: {MutationSpeed}",
                    $"Mutation Range: {MutationRange.X}, {MutationRange.Y}"
                })
            };

            for (int d = 1; d <= 365; d++)
            {
                dailyValues.Add(new DataLine());

                dailyValues[d].Values.Add(d.ToString());

                for (int e = 0; e < Evolution.Count; e++)
                {
                    dailyValues[d].Values.Add(Evolution[e].Chromosomes.First().BasePairs.Interpolate(d).Format(3));    
                }
            }

            dailyValues.Add(new DataLine(new List<string>() { "" }.Concat(Evolution.Select(e => e.FitNess.Format(3))).ToList()));
            dailyValues.Add(new DataLine(new List<string>() { "" }.Concat(Evolution.Select(e => e.Generation.Birth.ToString("HH:mm:ss"))).ToList()));

            new DataWriter().Write(fileName, dailyValues, ",");
        }
    }
}
