﻿using NMag.Core.Utilization;
using System;
using System.Collections.Generic;
using System.IO;

namespace NMag.Core.Simulation
{
    public class ResultReader
    {
        // First block
        private int _blockLength;
        private int _blockCount;
        private int _valuesPerBlock;
        private int _catchmentCount;
        private int _periodCount;
        private int _yearCount;
        private int _firstYear;
        private int _lastYear;
        private int _timeSteps;
        private int _firstStep;
        private int _lastStep;

        // Second block
        private int _catchmentCount2;

        public Result Read(SimulationResult simulationResult, string path)
        {
            try
            {
                using (FileStream stream = new FileStream(path, FileMode.Open))
                {
                    using (BinaryReader reader = new BinaryReader(stream))
                    {
                        UnpackFirstBlock(reader);

                        if (_catchmentCount != simulationResult.ModuleResults.Count)
                        {
                            return new Result(false, "Module count mismatch.");
                        }

                        UnpackSecondBlock(reader);
                        UnpackThirdBlock(reader);
                        UnpackForthBlock(reader);
                        simulationResult = UnpackFifthBlock(reader, simulationResult);

                        return new Result(true);
                    }
                }
            }
            catch
            {
                simulationResult = null;
                return new Result(false);
            }
        }

        private void UnpackFirstBlock(BinaryReader reader)
        {
            _blockLength = BitConverter.ToInt32(reader.ReadBytes(4), 0);
            _blockCount = BitConverter.ToInt32(reader.ReadBytes(4), 0);
            _valuesPerBlock = _blockLength / 4;

            _catchmentCount = (_blockLength - 20) / 40;
            _periodCount = (_blockCount - 5) / _catchmentCount;
            _yearCount = (_blockCount - 5) / 365;

            _firstYear = BitConverter.ToInt32(reader.ReadBytes(4), 0);
            _lastYear = BitConverter.ToInt32(reader.ReadBytes(4), 0);

            reader.ReadBytes(4);    // dummy

            _timeSteps = BitConverter.ToInt32(reader.ReadBytes(4), 0);
            _firstStep = BitConverter.ToInt32(reader.ReadBytes(4), 0);
            _lastStep = BitConverter.ToInt32(reader.ReadBytes(4), 0);

            reader.ReadBytes(4);    // parsim
            reader.ReadBytes((_valuesPerBlock - 9) * 4);    // dummy
        }

        private void UnpackSecondBlock(BinaryReader reader)
        {
            reader.ReadBytes(4);    // dummy
            _catchmentCount2 = BitConverter.ToInt32(reader.ReadBytes(4), 0);
            reader.ReadBytes(4);    // dummy
            reader.ReadBytes(4);    // nibyt
            reader.ReadBytes(4);    // nrbyt
            reader.ReadBytes(4);    // dum21
            reader.ReadBytes((_valuesPerBlock - 6) * 4);    // dumval
        }

        private void UnpackThirdBlock(BinaryReader reader)
        {
            reader.ReadBytes(_valuesPerBlock * 4);    // dummy
        }

        private List<double> UnpackForthBlock(BinaryReader reader)
        {
            List<double> catchmentMeanDischarges = new List<double>();

            for (int c = 0; c < _catchmentCount2; c++)
            {
                catchmentMeanDischarges.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));
            }

            reader.ReadBytes((_valuesPerBlock - _catchmentCount2) * 4);    // dummy

            return catchmentMeanDischarges;
        }

        private SimulationResult UnpackFifthBlock(BinaryReader reader, SimulationResult simulationResult)
        {
            reader.ReadBytes(_valuesPerBlock * 4);    // dummy

            for (int d = 0; d < _yearCount * 365; d++)
            {
                int year = BitConverter.ToInt32(reader.ReadBytes(4), 0);
                int day = BitConverter.ToInt32(reader.ReadBytes(4), 0);

                simulationResult.Periods.Add(DateHelper.DayToDate(year, day));
                simulationResult.FirmPowerProduction.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));
                simulationResult.TotalProduction.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));
                simulationResult.Cost.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));

                for (int m = 0; m < _catchmentCount; m++)
                {
                    simulationResult.ModuleResults[m].Storage.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));
                    simulationResult.ModuleResults[m].WaterLevel.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));

                    simulationResult.ModuleResults[m].LocalInflow.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));
                    simulationResult.ModuleResults[m].TotalInflow.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));

                    simulationResult.ModuleResults[m].ThroughFlow.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));
                    simulationResult.ModuleResults[m].Bypass.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));
                    simulationResult.ModuleResults[m].Spill.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));

                    simulationResult.ModuleResults[m].Production.Add(BitConverter.ToSingle(reader.ReadBytes(4), 0));

                    // Dummy
                    BitConverter.ToSingle(reader.ReadBytes(4), 0);
                    BitConverter.ToSingle(reader.ReadBytes(4), 0);
                }
            }

            return simulationResult;
        }
    }
}
