﻿using NMag.Core.Utilization;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;

namespace NMag.Core.Simulation
{
    public class ResultWriter
    {
        private const int MODULE_NAME_WIDTH = 40;

        public List<DataLine> FormatOverview(SimulationResult results)
        {
            if(results == null) { return new List<DataLine>(); }

            DataLine moduleNames = new DataLine(Enumerable.Repeat("", 4).ToList()
                .Concat(results.ModuleResults
                    .SelectMany(m => new List<string>() { ModuleHelper.GetModuleAddress(m.Module) }
                        .Concat(Enumerable.Repeat("", 7))
                        .ToList())
                    .ToList())
                .ToList());

            List<string> headers = new List<string>()
            {
                "",
                "FirmP",
                "TotP",
                "Cost"
            };
            
            List<string> units = new List<string>()
            {
                "",
                "GWh",
                "GWh",
                ""
            };

            List<DataLine> output = new List<DataLine>() { new DataLine(headers), new DataLine(units) };
            output.Add(new DataLine(new List<string>()
            {
                "Total",
                results.FirmPowerProductionSum.Format(3),
                results.TotalProductionSum.Format(3),
                results.CostSum.Format(3),
            }));
            output.Add(new DataLine(new List<string>()
            {
                "Average",
                results.FirmPowerProductionAverage.Format(3),
                results.TotalProductionAverage.Format(3),
                results.CostAverage.Format(3),
            }));
            output.Add(new DataLine());

            headers = new List<string>()
            {
                "Total".AddTail(' ', MODULE_NAME_WIDTH - 5),
                "InFLo",
                "TotFlo",
                "ThrFlo",
                "Bypass",
                "Spill",
                "Prod"
            };

            units = new List<string>()
            {
                "".AddTail(' ', MODULE_NAME_WIDTH),
                "mill m3",
                "mill m3",
                "mill m3",
                "mill m3",
                "mill m3",
                "MW"
            };
            output.Add(new DataLine());

            output.Add(new DataLine(headers));
            output.Add(new DataLine(units));
            foreach(ModuleResult moduleResult in results.ModuleResults)
            {
                string moduleAddress = ModuleHelper.GetModuleAddress(moduleResult.Module);

                output.Add(new DataLine(new List<string>()
                {
                    moduleAddress.AddTail(' ', MODULE_NAME_WIDTH - moduleAddress.Length),
                    moduleResult.LocalInflowSum.Format(3),
                    moduleResult.TotalInflowSum.Format(3),
                    moduleResult.ThroughFlowSum.Format(3),
                    moduleResult.BypassSum.Format(3),
                    moduleResult.SpillSum.Format(3),
                    moduleResult.ProductionSum.Format(3)
                }));
            }

            headers = new List<string>()
            {
                "Average".AddTail(' ', MODULE_NAME_WIDTH - 7),
                "InFLo",
                "TotFlo",
                "ThrFlo",
                "Bypass",
                "Spill",
                "Prod"
            };

            units = new List<string>()
            {
                "".AddTail(' ', MODULE_NAME_WIDTH),
                "mill m3",
                "mill m3",
                "mill m3",
                "mill m3",
                "mill m3",
                "MW"
            };
            output.Add(new DataLine());

            output.Add(new DataLine(headers));
            output.Add(new DataLine(units));
            foreach (ModuleResult moduleResult in results.ModuleResults)
            {
                string moduleAddress = ModuleHelper.GetModuleAddress(moduleResult.Module);

                output.Add(new DataLine(new List<string>()
                {
                    moduleAddress.AddTail(' ', MODULE_NAME_WIDTH - moduleAddress.Length),
                    moduleResult.LocalInflowAverage.Format(3),
                    moduleResult.TotalInflowAverage.Format(3),
                    moduleResult.ThroughFlowAverage.Format(3),
                    moduleResult.BypassAverage.Format(3),
                    moduleResult.SpillAverage.Format(3),
                    moduleResult.ProductionAverage.Format(3)
                }));
            }
            output.Add(new DataLine());

            return output;
        }

        public List<DataLine> FormatDetails(SimulationResult results)
        {
            if (results == null) { return new List<DataLine>(); }

            DataLine moduleNames = new DataLine(Enumerable.Repeat("", 4).ToList()
                .Concat(results.ModuleResults
                    .SelectMany(m => new List<string>() { ModuleHelper.GetModuleAddress(m.Module).AddTail(' ', 10 * 8 - ModuleHelper.GetModuleAddress(m.Module).Length) }
                        .ToList())
                    .ToList())
                .ToList());

            List<string> headers = new List<string>()
            {
                "Date",
                "FirmP",
                "TotP",
                "Cost"
            };

            for (int m = 0; m < results.ModuleResults.Count; m++)
            {
                headers.AddRange(new List<string>()
                {
                    "Stor",
                    "WLev",
                    "InFLo",
                    "TotFlo",
                    "ThrFlo",
                    "Bypass",
                    "Spill",
                    "Prod"
                });
            }

            List<string> units = new List<string>()
            {
                "Day",
                "GWh",
                "GWh",
                ""
            };
            for (int m = 0; m < results.ModuleResults.Count; m++)
            {
                units.AddRange(new List<string>()
                {
                    "mill m3",
                    "masl",
                    "m3 s-1",
                    "m3 s-1",
                    "m3 s-1",
                    "m3 s-1",
                    "m3 s-1",
                    "MW"
                });
            }

            List<DataLine> output = new List<DataLine>() { moduleNames, new DataLine(headers), new DataLine(units) };

            List<string> values = new List<string>();

            for (int d = 0; d < results.Periods.Count; d++)
            {
                values = new List<string>()
                {
                    results.Periods[d].ToString("yyyy-MM-dd"),
                    results.FirmPowerProduction[d].Format(3),
                    results.TotalProduction[d].Format(3),
                    results.Cost[d].Format(3),
                };

                for (int m = 0; m < results.ModuleResults.Count; m++)
                {
                    values.AddRange(new List<string>()
                    {
                        results.ModuleResults[m].Storage[d].Format(3),
                        results.ModuleResults[m].WaterLevel[d].Format(3),
                        results.ModuleResults[m].LocalInflow[d].Format(3),
                        results.ModuleResults[m].TotalInflow[d].Format(3),
                        results.ModuleResults[m].ThroughFlow[d].Format(3),
                        results.ModuleResults[m].Bypass[d].Format(3),
                        results.ModuleResults[m].Spill[d].Format(3),
                        results.ModuleResults[m].Production[d].Format(3)
                    });
                }

                output.Add(new DataLine(values));
            }
            output.Add(new DataLine());

            return output;
        }

        public void Write(SimulationResult results, string projectName, string simulationName, string name = "", bool useTab = false, bool overview = true, bool details = true)
        {
            if (results == null) { return; }

            string delimiter = ",";
            if(useTab)
            {
                delimiter = "\t";
            }

            List<string> output = new List<string>();
            if (overview)
            {
                output.AddRange(FormatOverview(results).Select(r => r.ToString(delimiter)));
            }
            if (overview)
            {
                output.AddRange(FormatDetails(results).Select(r => r.ToString(delimiter)));
            }

            string simulationFolder = Path.Combine(PathFinder.GetFolder(PathFinder.Folder.Project, projectName), simulationName == "" ? "Results" : simulationName);
            if (!Directory.Exists(simulationFolder)) { Directory.CreateDirectory(simulationFolder); }
            string fileName = "";
            if(name == "")
            {
                fileName = Path.Combine(simulationFolder, results.ProjectName + " results" + ".csv");
            }
            else
            {
                fileName = Path.Combine(simulationFolder, name + ".csv");
            }


            new DataWriter().Write(fileName, output, ",");
        }
    }
}
