﻿using NMag.Core;
using NMag.Core.Utilization;
using NMag.UI.CustomControls;
using NMag.UI.Dialogs;
using NMag.UI.Windows;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Windows;
using System;

namespace NMag.UI.Views
{
    public class HydrologyViewModel : MainViewModel
    {
        private RunoffSeries _selectedSeries;

        private string _scalingFactor;

        private string _seriesFirstYear;
        private string _seriesLastYear;

        public HydrologyViewModel(ApplicationViewModel application)
            :base(application, "Hydrological Data", ViewModelGroup.Setup)
        {
            SelectedName = RunoffSeries.FirstOrDefault() ?? "";

            GetData();
        }

        public override ToolbarViewModel Toolbar
        {
            get { return new HydrologyToolbarViewModel(this); }
        }

        //public string FirstYear
        //{
        //    get
        //    {
        //        return (Hydrology?.FirstYear ?? 0) == 0 ?
        //            "" :
        //            string.Format($"First Year: {Hydrology.FirstYear}");
        //    }
        //}
        //public string LastYear
        //{
        //    get
        //    {
        //        return (Hydrology?.LastYear ?? 0) == 0 ?
        //            "" :
        //            string.Format($"Last Year: {Hydrology.LastYear}");
        //    }
        //}
        public string TimeResolution // TODO
        {
            get { return string.Format($"Time Resolution: {Hydrology?.TimeResolution ?? 0} periods per year"); }
        }
        public string Compatibility
        {
            get { return Hydrology?.Validate().Message ?? ""; }
        }

        public List<string> RunoffSeries
        {
            get { return Hydrology?.RunoffSeries.Select(r => r.Name).ToList() ?? new List<string>(); }
        }
        public string SelectedName
        {
            get { return _selectedSeries?.Name ?? ""; }
            set
            {
                SaveData();
                _selectedSeries = Hydrology?.RunoffSeries.FirstOrDefault(r => r.Name == value);
                GetData();
                Update();
            }
        }

        public bool IsSeriesSelected
        {
            get { return SelectedName != ""; }
        }

        public string SeriesLastYear
        {
            get { return _seriesLastYear; }
            set
            {
                _seriesLastYear = value;
                SaveData();
                Update();
            }
        }
        public string SeriesFirstYear
        {
            get { return _seriesFirstYear; }
            set
            {
                _seriesFirstYear = value;
                SaveData();
                Update();
            }
        }
        public string StationNumber
        {
            get { return _selectedSeries?.StationNumber.ToString() ?? ""; }
            set
            {
                if (_selectedSeries != null && int.TryParse(value, out int number))
                {
                    _selectedSeries.StationNumber = number;
                }
            }
        }
        public string StationCode
        {
            get { return _selectedSeries?.StationCode ?? ""; }
            set
            {
                if(_selectedSeries != null)
                {
                    _selectedSeries.StationCode = value;
                }
            }
        }
        public string ScalingFactor
        {
            get { return _scalingFactor; }
            set { _scalingFactor = value; }
        }
        public string DataSource
        {
            get { return _selectedSeries?.HydroCurve.DataSource ?? ""; }
        }

        public ChartViewModel HydroChart
        {
            get
            {
                if (_selectedSeries?.HydroCurve == null) { return null; }

                return new ChartViewModel(
                    new SeriesData(
                        _selectedSeries?.HydroCurve, 
                        SeriesData.SeriesType.Line),
                    Update);
            }
        }

        private Hydrology Hydrology
        {
            get { return Application.Project?.Hydrology; }
        }

        public void AddSeries()
        {
            if(Hydrology == null) { return; }

            if(Hydrology.RunoffSeries.Count == 3)
            {
                MessageBox.Show("Curve limit of 3 reached.", "Curve Limit Reached", MessageBoxButton.OK, MessageBoxImage.Error);
                return;
            }

            SeriesData hydroData = new SeriesData(
                new HydroCurve(
                    new List<Core.Point>(),
                    0),
                SeriesData.SeriesType.Line);

            DataViewerWindow dataView = new DataViewerWindow(hydroData, Update);
            DataViewerViewModel dataViewModel = ((DataViewerViewModel)dataView.DataContext);

            dataViewModel.ImportData();
            if(dataViewModel.Curve == null ? false : dataViewModel.Curve.Points.Count == 0)
            {
                return;
            }

            dataView.ShowDialog();

            Curve curve = dataViewModel.Curve;

            if(curve.Points.Count == 0 || curve.FirstYear == 0) { return; }

            string name = StringHelper.GetUniqueName(
                Path.GetFileNameWithoutExtension(curve.DataSource), 
                Hydrology.RunoffSeries.Select(r => r.Name).ToList());

            Hydrology.AddRunoffSeries(new RunoffSeries(name) { HydroCurve = new HydroCurve(curve.Points.Select(p => p.Clone()).ToList(), curve.FirstYear) });
            Application.Project.Modules.ForEach(m => m.AddRunoffSeries(name));

            OnPropertyChanged("RunoffSeries");
            SelectedName = Hydrology.RunoffSeries.Last().Name;
        }

        public void EditSeries()
        {
            if(_selectedSeries == null) { return; }
            
            TextInputDialog dialog = new TextInputDialog("Series Name", "What is the name of this runoff series?", "No name given!", SelectedName);
            if (dialog.ShowDialog() == true)
            {
                if (dialog.Result == SelectedName) { return; }

                if (Hydrology.RunoffSeries.Count(s => s.Name == dialog.Result) > 0)
                {
                    MessageBox.Show("A series with this name already exists.", "Duplicate Name", MessageBoxButton.OK, MessageBoxImage.Exclamation);
                    return;
                }

                Application.Project.Modules.ForEach(m => m.RenameRunoffSeries(_selectedSeries.Name, dialog.Result));
                Hydrology.EditRunoffSeries(_selectedSeries.Name, dialog.Result);

                OnPropertyChanged("RunoffSeries");
                SelectedName = dialog.Result;
            }
        }

        public void DeleteSeries()
        {
            if(_selectedSeries == null || Hydrology == null) { return; }

            if(MessageBox.Show("Are you sure you want to delete this series?\nThis action cannot be undone.", "Delete Series", MessageBoxButton.YesNo, MessageBoxImage.Exclamation) == MessageBoxResult.Yes)
            {
                Hydrology.RemoveRunoffSeries(_selectedSeries.Name);
                Application.Project.Modules.ForEach(m => m.RemoveRunoffSeries(_selectedSeries.Name));

                OnPropertyChanged("RunoffSeries");
                SelectedName = Hydrology.RunoffSeries.LastOrDefault()?.Name;
            }
        }

        public override void Update()
        {
            UpdateToolbar();

            //OnPropertyChanged("FirstYear");
            //OnPropertyChanged("LastYear");
            OnPropertyChanged("Compatibility");

            OnPropertyChanged("SelectedName");
            OnPropertyChanged("IsSeriesSelected");

            OnPropertyChanged("SeriesFirstYear");
            OnPropertyChanged("SeriesLastYear");
            OnPropertyChanged("StationNumber");
            OnPropertyChanged("StationCode");
            OnPropertyChanged("ScalingFactor");
            OnPropertyChanged("DataSource");

            OnPropertyChanged("HydroChart");
        }

        public override void SaveData()
        {
            if (_selectedSeries == null) { return; }

            if(double.TryParse(ScalingFactor, out double value)) { _selectedSeries.ScalingFactor = value; }
            if (int.TryParse(SeriesFirstYear, out int year)) { _selectedSeries.HydroCurve.FirstYear = year; }
            if (int.TryParse(SeriesLastYear, out year)) { _selectedSeries.HydroCurve.LastYear = year; }
        }

        public override void GetData()
        {
            if (_selectedSeries == null) { return; }

            ScalingFactor = _selectedSeries.ScalingFactor.ToString();
            SeriesFirstYear = _selectedSeries.HydroCurve.FirstYear.ToString();
            SeriesLastYear = _selectedSeries.HydroCurve.LastYear.ToString();
        }
    }
}
