﻿using NMag.Core;
using NMag.Core.Optimization;
using NMag.Core.Simulation;
using NMag.UI.Miscellaneous;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Linq;
using System.Threading;
using System.Windows.Input;

namespace NMag.UI.Views
{
    public class SimulationViewModel : MainViewModel
    {
        private string _firstYear;
        private string _lastYear;
        private string _reservoirLevel;
        private string _firmPowerLevel;

        private string _generationCount = "30";
        private string _populationSize = "50";

        private string _geneComplexity = "10";

        private string _mutationProbability = "90";
        private string _mutationSpeed = "20";
        private string _mutationRange = "2";

        private Optimizer _optimizer;

        private object _lock = new object();
        private static List<string> _status = new List<string>();

        public SimulationViewModel(ApplicationViewModel application)
            : base(application, "Simulation", ViewModelGroup.Simulation)
        {
            GetData();
        }

        public override ToolbarViewModel Toolbar
        {
            get { return null; }
        }

        public string FirstYear
        {
            get { return _firstYear; }
            set { _firstYear = value; }
        }
        public string LastYear
        {
            get { return _lastYear; }
            set { _lastYear = value; }
        }
        public string ReservoirLevel
        {
            get { return _reservoirLevel; }
            set { _reservoirLevel = value; }
        }
        public string FirmPowerLevel
        {
            get { return _firmPowerLevel; }
            set { _firmPowerLevel = value; }
        }

        public string SimulationName { get; set; } = DateTime.Now.ToString("yyyyMMddHHmmss");

        public string GenerationCount
        {
            get { return _generationCount; }
            set { _generationCount = value; }
        }
        public string PopulationSize
        {
            get { return _populationSize; }
            set { _populationSize = value; }
        }

        public string GeneComplexity
        {
            get { return _geneComplexity; }
            set { _geneComplexity = value; }
        }

        public string MutationProbability
        {
            get { return _mutationProbability; }
            set { _mutationProbability = value; }
        }
        public string MutationSpeed
        {
            get { return _mutationSpeed; }
            set { _mutationSpeed = value; }
        }
        public string MutationRange
        {
            get { return _mutationRange; }
            set { _mutationRange = value; }
        }

        public string Validation
        {
            get { return Application.Project?.Validate().Message ?? ""; }
        }

        public ObservableCollection<string> Status
        {
            get
            {
                Monitor.Enter(_status);
                try
                {
                    return new ObservableCollection<string>(GetStatus());
                }
                finally
                {
                    Monitor.Exit(_status);
                }            
            }
        }

        public int SelectedStatus
        {
            get { return Status.Count - 1; }
        }

        public ICommand RunSimulationCommand
        {
            get
            {
                return new RelayCommand(args => RunSimulation(),
                    args => Application.Project != null &&
                        Application.Project.Validate().Succes &&
                        (!_optimizer?.IsRunning ?? true));
            }
        }

        public ICommand StartOptimizerCommand
        {
            get
            {
                return new RelayCommand(args => StartOptimizer(),
                    args => Application.Project != null &&
                        Application.Project.Validate().Succes &&
                        (!_optimizer?.IsRunning ?? true));
            }
        }

        public ICommand StopOptimizerCommand
        {
            get
            {
                return new RelayCommand(args => StopOptimizer(),
                    args => _optimizer?.IsRunning ?? false);
            }
        }

        public ICommand ClearStatusCommand
        {
            get
            {
                return new RelayCommand(args => 
                {
                    _status.Clear();
                    OnPropertyChanged("Status");
                    OnPropertyChanged("SelectedStatus");
                });
            }
        }

        public override void GetData()
        {
            if(Application.Project?.JobControl == null) { return; }

            FirstYear = Application.Project.JobControl.FirstYear.ToString();
            LastYear = Application.Project.JobControl.LastYear.ToString();
            FirmPowerLevel = Application.Project.JobControl.FirmPowerLevel.ToString();
            ReservoirLevel = Application.Project.JobControl.ReservoirLevel.ToString();
        }

        public override void SaveData()
        {
            if (Application.Project?.JobControl == null) { return; }

            if(int.TryParse(FirstYear, out int intValue)) { Application.Project.JobControl.FirstYear = intValue; }
            if (int.TryParse(LastYear, out intValue)) { Application.Project.JobControl.LastYear = intValue; }
            if (int.TryParse(FirmPowerLevel, out intValue)) { Application.Project.JobControl.FirmPowerLevel = intValue; }
            if (double.TryParse(ReservoirLevel, out double doubleValue)) { Application.Project.JobControl.ReservoirLevel = doubleValue; }
        }

        public override void Update()
        {
            UpdateToolbar();

            OnPropertyChanged("FirstYear");
            OnPropertyChanged("LastYear");
            OnPropertyChanged("ReservoirLevel");
            OnPropertyChanged("FirmPowerLevel");

            OnPropertyChanged("Validation");
            OnPropertyChanged("Status");
            OnPropertyChanged("SelectedStatus");
        }

        private void RunSimulation()
        {
            SaveData();
            Report("Start Simulation...");

            Simulation simulation = new Simulation(SimulationName);
            SimulationResult result = simulation.Run(Application.Project, ClearLevel.All);

            if(result != null)
            {
                Application.SimulationResult = result;
                Report("Simulation Completed!");
            }
            else
            {
                foreach(string error in simulation.Errors)
                {
                    Report(error);
                }
            }
        }

        private void StartOptimizer()
        {
            SaveData();

            int.TryParse(GenerationCount, out int generationCount);
            int.TryParse(PopulationSize, out int populationSize);
            int.TryParse(GeneComplexity, out int geneComplexity);
            int.TryParse(MutationProbability, out int mutationProbability);
            int.TryParse(MutationSpeed, out int mutationSpeed);
            int.TryParse(MutationRange, out int range);
            Point mutationRange = new Point(range, range);

            PopulationControl populationControl = new PopulationControl(generationCount, populationSize, geneComplexity, mutationProbability, mutationSpeed, mutationRange);

            _optimizer = new Optimizer(Report, SimulationName);
            _optimizer.Start(Application.Project.Clone(), populationControl);
        }

        private void StopOptimizer()
        {
            _optimizer.Abort();
        }

        private List<string> GetStatus(string line = "")
        {
            lock(_lock)
            {
                if(line != "")
                {
                    _status.Add(line);
                }

                List<string> status = new List<string>(_status);
                return status;
            }
        }

        private void Report(string line)
        {
            GetStatus(DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss ") + line);
            OnPropertyChanged("Status");
            OnPropertyChanged("SelectedStatus");
        }
    }
}