﻿using NMag.Core;
using NMag.UI.Miscellaneous;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.Data;
using System.IO;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;

using Forms = System.Windows.Forms;

namespace NMag.UI.Windows
{
    public class DataImporterViewModel : WindowViewModel
    {
        private DataReader _reader = new DataReader();
        private Curve _curve;

        private List<Core.Point> points = new List<Core.Point>();

        private string _xColumn = "";
        private string _yColumn = "";

        public DataImporterViewModel(EventHandler requestCloseHandler, ref Curve curve)
            : base (requestCloseHandler)
        {
            _curve = curve;
        }

        public string XAxisTitle
        {
            get
            {
                return _curve.Axes[0].Title + ":";
            }
        }
        public string YAxisTitle
        {
            get
            {
                return _curve.Axes[1].Title + ":";
            }
        }

        public ObservableCollection<DataGridColumn> ColumnCollection
        {
            get
            {
                return new ObservableCollection<DataGridColumn>(GetColumns(_reader.Data));
            }
        }

        public List<string> ColumnHeaders
        {
            get
            {
                return GetColumns(_reader.Data).Select(c => c.Header.ToString()).ToList();
            }
        }

        public string XColumn
        {
            get
            {
                return _xColumn;
            }
            set
            {
                _xColumn = value;
                OnPropertyChanged("XColumn");
            }
        }
        public string YColumn
        {
            get
            {
                return _yColumn;
            }
            set
            {
                _yColumn = value;
                OnPropertyChanged("YColumn");
            }
        }

        public DataView Data
        {
            get
            {
                return _reader.Data?.DefaultView;
            }
        }

        public ICommand LoadFileCommand
        {
            get
            {
                return new RelayCommand(args => LoadFile());
            }
        }
        public ICommand ImportDataCommand
        {
            get
            {
                return new RelayCommand(args => ImportData(),
                    args => (XColumn != "") && (YColumn != ""));
            }
        }

        private IEnumerable<DataGridColumn> GetColumns(DataTable table)
        {
            if(table != null)
            {
                foreach(DataColumn column in table.Columns)
                {
                    yield return new DataGridTextColumn()
                    {
                        Header = column.ColumnName,
                        IsReadOnly = true,
                        Binding = new System.Windows.Data.Binding(column.ColumnName)
                    };
                }
            }
        }

        private void LoadFile()
        {
            Forms.OpenFileDialog dialog = new Forms.OpenFileDialog()
            {
                Multiselect = false,
                Filter = GetFileFilter()
            };

            if(dialog.ShowDialog() == Forms.DialogResult.OK)
            {
                string extention = Path.GetExtension(dialog.FileName);

                Result result = new Result(true);

                switch (extention)
                {
                    case ".xlsx":
                    case ".xlsm":
                    case ".xls":
                        result = _reader.ReadExcel(dialog.FileName);
                        break;

                    case ".txt":
                    case ".csv":
                        result = _reader.ReadText(dialog.FileName);
                        break;

                    default:
                        Forms.MessageBox.Show($"Cannot load data from file with extention \"{extention}\".", "Invalid File Type", Forms.MessageBoxButtons.OK, Forms.MessageBoxIcon.Error);
                        return;
                }

                if (!result.Succes)
                {
                    MessageBox.Show(result.Message, "Reading Failed", MessageBoxButton.OK, MessageBoxImage.Error);
                    return;
                }

                Update();
            }
        }

        public override void Update()
        {
            XColumn = "";
            YColumn = "";
            OnPropertyChanged("ColumnCollection");
            OnPropertyChanged("ColumnHeaders");
            OnPropertyChanged("Data");
        }
        
        private string GetFileFilter()
        {
            string filter = "";

            if (Type.GetTypeFromProgID("Excel.Application") != null)
            {
                filter = string.Join("|", filter, "Excel Files (*.xlsx;*.xlsm;*.xls)|*.xlsx;*.xlsm;*.xls");
            }

            filter = string.Join("|", filter, "Delimited Files (*.csv)|*.csv");
            filter = string.Join("|", filter, "Text Files (*.txt)|*.txt");

            return filter.Remove(0, 1);
        }
        
        private void ImportData()
        {
            if (_reader.GetData(XColumn, YColumn, out List<Core.Point> points, out string dataSource))
            {
                _curve.Points = points;

                if (_curve.Type == CurveType.Hydro)
                {
                    _curve.DataSource = dataSource;
                    _curve.FirstYear = _reader.FirstYear;
                }
            }

            Close();
        }
    }
}
