﻿using NMag.Core;
using NMag.UI.CustomControls;
using NMag.UI.Dialogs;
using NMag.UI.Miscellaneous;
using System;
using System.Collections.Generic;
using System.Collections.ObjectModel;
using System.ComponentModel;
using System.Linq;
using System.Windows;
using System.Windows.Input;

namespace NMag.UI.Windows
{
    public class DataViewerViewModel : WindowViewModel
    {
        private Action _updateAction;

        private Curve _curve;
        private List<int> _selectedPoints = new List<int>();

        public DataViewerViewModel(EventHandler requestCloseHandler, SeriesData seriesData, Action updateAction)
            : base (requestCloseHandler)
        {
            SeriesData = seriesData;
            _curve = seriesData.Curve.Clone();
            
            _updateAction = updateAction;

            Update(true);
        }

        public SeriesData SeriesData { get; private set; }
        public Curve Curve
        {
            get { return _curve; }
        }

        public ObservableCollection<Core.Point> Points
        {
            get
            {
                return new ObservableCollection<Core.Point>(_curve.Points);
            }
        }
        public List<int> SelectedPoints
        {
            get
            {
                return _selectedPoints;
            }
            set
            {
                _selectedPoints = value;
            }
        }

        public string InvalidText
        {
            get
            {
                return _curve.Validate().Message;
            }
        }
        public bool IsDataValid
        {
            get
            {
                return _curve.Validate().Succes;
            }
        }

        public string XAxisTitle
        {
            get
            {
                return _curve.Axes[0].Title;
            }
        }
        public string YAxisTitle
        {
            get
            {
                return _curve.Axes[1].Title;
            }
        }

        public string MaximumRowCount
        {
            get
            {
                return (_curve.MaxLength == 0) ?
                    "" : 
                    $"Maximum number of rows: {_curve.MaxLength}";
            }
        }
        public bool ShowDates
        {
            get
            {
                return _curve.Axes[0].TimeScale != TimeScale.None;
            }
        }
        public int XPrecision
        {
            get
            {
                return _curve.Axes[0].DecimalValues ? 3 : 0;
            }
        }

        public int FirstYear
        {
            get
            {
                return _curve.FirstYear;
            }
        }

        public ChartViewModel ChartViewModel
        {
            get
            {
                return new ChartViewModel(new SeriesData(_curve, SeriesData.Type));
            }
        }

        public ICommand UpdateDataCommand
        {
            get
            {
                return new RelayCommand(args => Update(false));
            }
        }

        public ICommand ImportDataCommand
        {
            get
            {
                return new RelayCommand(args => ImportData());
            }
        }
        public ICommand ExportDataCommand
        {
            get
            {
                return new RelayCommand(args => ExportData(),
                    args => (_curve != null) && (_curve.Points.Count > 0));
            }
        }

        public ICommand DeleteRowsCommand
        {
            get
            {
                return new RelayCommand(args => DeleteRows(SelectedPoints?.ToList()),
                    args => SelectedPoints.Count > 0);
            }
        }
        public ICommand InsertRowCommand
        {
            get
            {
                return new RelayCommand(args => InsertRows(SelectedPoints.FirstOrDefault(), 1),
                    args => _curve.MaxLength != 0 && 
                        _curve.Points.Count < _curve.MaxLength);
            }
        }
        public ICommand InsertRowsCommand
        {
            get
            {
                return new RelayCommand(args =>
                {
                    string description = "How many rows should be added?";

                    if(_curve.MaxLength > 0)
                    {
                        description += $" (Max: {_curve.MaxLength - _curve.Points.Count})";
                    }

                    NumberInputDialog dialog = new NumberInputDialog("Insert Rows", description, 0, 1);
                    dialog.ShowDialog();
                    
                    InsertRows(SelectedPoints.Count == 0 ? _curve.Points.Count : SelectedPoints.First(), (int)dialog.Result);
                },
                args => _curve.MaxLength != 0 &&
                    _curve.Points.Count < _curve.MaxLength);
            }
        }

        public void OnWindowClosing(object sender, CancelEventArgs e)
        {
            Update(false);

            if(IsDataValid)
            {
                SeriesData.Curve.Points = _curve.Points;
                SeriesData.Curve.FirstYear = _curve.FirstYear;
                _updateAction?.Invoke();
            }
        }

        public void ImportData()
        {
            DataImporterView importer = new DataImporterView(_curve);
            importer.ShowDialog();

            Update(true);
        }

        public override void Update()
        {
            Update(true);
        }

        private void Update(bool rowCountChanged)
        {
            if(rowCountChanged)
            {
                OnPropertyChanged("Points");
            }

            if (IsDataValid)
            {
                OnPropertyChanged("ChartViewModel");
            }

            OnPropertyChanged("InvalidText");
            OnPropertyChanged("IsDataValid");
        }

        private void ExportData()
        {
            DataExporterView exporter = new DataExporterView(_curve);
            exporter.ShowDialog();
        }

        private void DeleteRows(List<int> rowIndices)
        {
            if(MessageBox.Show("Are you sure you want to delete the selected rows?\nThis action cannot be undone.", 
                "Delete selected rows", 
                MessageBoxButton.YesNo, 
                MessageBoxImage.Warning) 
                == MessageBoxResult.No)
            {
                return;
            }

            List<int> removeIndices = new List<int>(rowIndices);
            removeIndices.Reverse();

            foreach(int i in removeIndices)
            {
                _curve.Points.RemoveAt(i);
            }

            Update(true);
        }

        private void InsertRows(int index, int count)
        {
            count = Math.Min(count, _curve.MaxLength - _curve.Points.Count);

            for (int i = 0; i < count; i++)
            {
                _curve.Points.Insert(index + i, new Core.Point(0, 0));
            }

            Update(true);
        }
    }
}
